package hu.uw.pallergabor.peeble;

import android.content.ContentResolver;
import android.content.ContentValues;
import android.database.Cursor;
import android.net.Uri; 
import android.util.Log;

import mwfw.sync.IStorage;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Map;

/**
  * Adapter class mapping MWFW Sync framework storage abstraction to PeebleDataProvider
  * calls. Serves the LocDistro table.
  */

public class LocDistroSyncStorageAdapter implements IStorage {
  private ContentResolver cr = null;
  private String lockKey = null;
  private static final String LOG_TAG = "Peeble_LocDistroSyncStorageAdapter";

  public LocDistroSyncStorageAdapter( ContentResolver cr ) {
	this.cr = cr;
  }

  public void init() {}
  public void shutdown() {}

/**
  * Gets synchronization metainfo associated with the database as collection of key-value pairs.
  * The definition of these key-value pairs depend on the synchronization engine. 
  * @return the synchronization metainformation associated with the database
  */
  public Map getDatabaseMetaInfo() {
	HashMap<String,String> m = new HashMap<String,String>();
	String columns[] = { PeebleData.SyncMetaInfo.NAME,
						PeebleData.SyncMetaInfo.VALUE };
	Cursor c = cr.query( PeebleData.LocDistro.SYNCMETAINFO_CONTENT_URI,
				columns, null, null, null );
	if( c == null )
		return m;
	while( c.next() ) {
	  int nameColumnIndex = c.getColumnIndex( PeebleData.SyncMetaInfo.NAME );
	  int valueColumnIndex = c.getColumnIndex( PeebleData.SyncMetaInfo.VALUE );
	  String name = c.getString( nameColumnIndex );
	  String value = c.getString( valueColumnIndex );
	  m.put( name, value );
	}
	c.close();
	return m;
  }

/**
  * Sets synchronization metainfo associated with the database as collection of key-value pairs.
  * The definition of these key-value pairs depend on the synchronization engine. 
  * @param metainf 
  */
  public void setDatabaseMetaInfo( Map metainf ) {
	for( Iterator it = metainf.keySet().iterator() ; it.hasNext() ; ) {
	  String key = (String)it.next();
	  String value = (String)metainf.get( key );
	  String columns[] = {  PeebleData.SyncMetaInfo._ID,
							PeebleData.SyncMetaInfo.NAME,
							PeebleData.SyncMetaInfo.VALUE };
	  Cursor c = cr.query( PeebleData.LocDistro.SYNCMETAINFO_CONTENT_URI,
				columns, 
				"name=?", 
				new String[] { key }, 
				null );
	  if( ( c == null ) || !c.next() ) {
		if( c != null )
			c.close();
	    ContentValues val = new ContentValues();
	    val.put( PeebleData.SyncMetaInfo.NAME, key );
	    val.put( PeebleData.SyncMetaInfo.VALUE, value );
	    cr.insert( PeebleData.LocDistro.SYNCMETAINFO_CONTENT_URI,val );
	  } else {
	    int valueColumnIndex = c.getColumnIndex( PeebleData.SyncMetaInfo.VALUE );
		c.updateString( valueColumnIndex,value );
		c.commitUpdates();
	    c.close();
	  }
	}
  }

/**
  * Selects database keys according to certain criteria. The criteria is specified as an
  * LDAP query executed on metainfo keys and values
  * @param query map. If it is null. all keys in the database will be selected.
  * @return Iterator that can be used to enumerate the keys satisfying the criteria
  */
  public Iterator selectKeys( Map queryMap ) {
	Long timestampLimit = (Long)queryMap.get( "timestampgeq" );
	if( timestampLimit == null )
		return null;
	String columns[] = { PeebleData.LocDistro._ID };
	Uri queryUri = Uri.withAppendedPath( PeebleData.LocDistro.CONTENT_URI, "delete" );
	Cursor c = cr.query( queryUri,
				columns, 
				"_sync_version > ?", 
				new String[] { timestampLimit.toString() }, 
				null );
	if( c == null )
		return null;
	ArrayList<String> l = new ArrayList<String>();
	while( c.next() ) {
	  int idColumnIndex = c.getColumnIndex( PeebleData.SyncMetaInfo._ID );
	  String id = c.getString( idColumnIndex );
	  l.add( id );
	}
	c.close();
	return l.iterator();
  }  

/**
  * Gets the meta info map associated with a key
  * @param key Key in the database to query
  * @return Meta information in the form of key-value pairs or null if there's no meta information 
  *         associated with the entry.
  */
  public Map getEntryMetaInfo( String key ) {
	Uri queryUri = Uri.withAppendedPath( PeebleData.LocDistro.CONTENT_URI, "delete" );
	queryUri = Uri.withAppendedPath( queryUri,key );
	String columns[] = { PeebleData.LocDistro._ID,
						PeebleData.LocDistro._SYNC_VERSION,
						PeebleData.LocDistro._SYNC_DELETED,
						PeebleData.LocDistro._SYNC_CREATED };
	Cursor c = cr.query( queryUri, columns, null, null, null );
	if( c == null )
		return null;
	if( !c.next() ) {
	  c.close();
	  return null;
	}
	int idColumnIndex = c.getColumnIndex( PeebleData.LocDistro._ID );
	int syncVersionColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_VERSION );
	int syncDeletedColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_DELETED );
	int syncCreatedColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_CREATED );
	long idValue = c.getLong( idColumnIndex );
	long syncVersionValue = c.getLong( syncVersionColumnIndex );
	long syncDeletedValue = c.getLong( syncDeletedColumnIndex );
	long syncCreatedValue = c.getLong( syncCreatedColumnIndex );
	HashMap m = new HashMap();
	m.put( "timestamp", Long.toString( syncVersionValue ) );
	m.put( "created", Long.toString( syncCreatedValue ) );
	if( syncDeletedValue > 0L )
		m.put( "deleted",Long.toString( syncDeletedValue ) );
// MIME type is not used but the sync engine breaks if this metainfo entry is not present
// (because it is used by the SyncML protocol)
	m.put( "type","text/plain" );
	c.close();
	return m;
  }

/**
  * Gets the meta info map associated with a key
  * @param key Key in the database to query
  * @param metainf Meta info to associate with the key
  */
  public void setEntryMetaInfo( String key, Map metainf ) {
	Uri queryUri = Uri.withAppendedPath( PeebleData.LocDistro.CONTENT_URI, "delete" );
	queryUri = Uri.withAppendedPath( queryUri,key );
	String columns[] = { PeebleData.LocDistro._ID,
						PeebleData.LocDistro._SYNC_VERSION,
						PeebleData.LocDistro._SYNC_DELETED,
						PeebleData.LocDistro._SYNC_CREATED };
	Cursor c = cr.query( queryUri, columns, null, null, null );
	if( c == null )
	  return;
	if( !c.next() ) {
	  c.close();
	  return;
	}
    long deleted = -1;
    try {
      deleted = Long.parseLong( (String)metainf.get( "deleted" ) );
    } catch( Exception ex ) {}
    long created = -1;
    try {
      created = Long.parseLong( (String)metainf.get( "created" ) );
    } catch( Exception ex ) {}
    long timestamp = -1;
    try {
      timestamp = Long.parseLong( (String)metainf.get( "timestamp" ) );
    } catch( Exception ex ) {}
	int syncVersionColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_VERSION );
	int syncDeletedColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_DELETED );
	int syncCreatedColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_CREATED );
	boolean commitNeeded = false;
	if( deleted >= 0 ) {
	  c.updateLong( syncDeletedColumnIndex, deleted );
	  commitNeeded = true;
	}
	if( created >= 0 ) {
	  c.updateLong( syncCreatedColumnIndex, created );
	  commitNeeded = true;
	}
	if( timestamp >= 0 ) {
	  c.updateLong( syncVersionColumnIndex, timestamp );
	  commitNeeded = true;
	}
	if( commitNeeded )
		c.commitUpdates();
	c.close();
  }

/**
  * Gets the data of the entry.
  * Database entry format: 
  * <centerLongitude>,<centerLatitude>,<counter>
  * @param key Key to read
  * @return The entry data as InputStream. The entry data can be read from this stream. Returns
  *         null if the entry doesn't exist.
  *
  */
  public InputStream getEntryData( String key ) {
	Uri queryUri = Uri.withAppendedPath( PeebleData.LocDistro.CONTENT_URI, key );
	String columns[] = { PeebleData.LocDistro.LONGITUDE_CENTER,
						PeebleData.LocDistro.LATITUDE_CENTER,
						PeebleData.LocDistro.COUNTER };
	Cursor c = cr.query( queryUri, columns, null, null, null );
	if( c == null )
	  return null;
	if( !c.next() ) {
	  c.close();
	  return null;
	}
	int longitudeColumnIndex = c.getColumnIndex( PeebleData.LocDistro.LONGITUDE_CENTER );
	int latitudeColumnIndex = c.getColumnIndex( PeebleData.LocDistro.LATITUDE_CENTER );
	int counterColumnIndex = c.getColumnIndex( PeebleData.LocDistro.COUNTER );
	String longitude = c.getString( longitudeColumnIndex );
	String latitude = c.getString( latitudeColumnIndex );
	String counter = c.getString( counterColumnIndex );
	String fieldValue = longitude+","+latitude+","+counter;
    ByteArrayInputStream bis = null;
	try {
	  bis = new ByteArrayInputStream( fieldValue.getBytes( "UTF-8") );
	} catch( UnsupportedEncodingException ex ) {
	  return null;
	}
	c.close();
    return (InputStream)bis;
  }

/**
  * Sets the data and the metainfo of the entry
  * @param key Key to update
  * @param metainf The meta information to be associated with the entry
  * @param data InputStream that can be used to read the data from
  * @return true if the entry was created, false if it was modified
  */
  public boolean setEntryData( String key, Map metainf, InputStream data ) {
    long deleted = -1;
    long created = -1;
    long timestamp = -1;
	long ts = -1;
    try {
      deleted = Long.parseLong( (String)metainf.get( "deleted" ) );
    } catch( Exception ex ) {}
    try {
      created = Long.parseLong( (String)metainf.get( "created" ) );
    } catch( Exception ex ) {}
    try {
      timestamp = Long.parseLong( (String)metainf.get( "timestamp" ) );
    } catch( Exception ex ) {}
	if( metainf.get( "autostamp" ) != null )
	  ts = System.currentTimeMillis();
    StringBuffer b = new StringBuffer();
    int ch = 0;
	try {
      do {
        ch = data.read();
        if( ch >= 0 )
          b.append( (char)ch );
      } while( ch >= 0 );
	} catch( IOException ex ) {
	  return false;
	}
    String s = new String( b );
	String sa[] = s.split( "," );
	if( sa.length < 3 )
		return false;
	Uri queryUri = Uri.withAppendedPath( PeebleData.LocDistro.CONTENT_URI, "location" );
	queryUri = Uri.withAppendedPath( queryUri, sa[0]+","+sa[1] );
	String columns[] = {
	  PeebleData.LocDistro._ID,
	  PeebleData.LocDistro.LONGITUDE_CENTER,
	  PeebleData.LocDistro.LATITUDE_CENTER,
	  PeebleData.LocDistro.COUNTER,
	  PeebleData.LocDistro._SYNC_DELETED,
	  PeebleData.LocDistro._SYNC_CREATED,
	  PeebleData.LocDistro._SYNC_VERSION
	};
	boolean resultValue = false;
	Cursor c = cr.query( queryUri, columns, null, null, null );
	if( ( c == null ) || !c.next() ) {
// coordinate bucket does not exist, we insert
	  if( c != null )
		c.close();
	  ContentValues val = new ContentValues();
	  val.put( PeebleData.LocDistro.LONGITUDE_CENTER, sa[0] );
	  val.put( PeebleData.LocDistro.LATITUDE_CENTER, sa[1] );
	  val.put( PeebleData.LocDistro.COUNTER,sa[2] );
// if autostamped, set the "created" and "timestamp" fields to the current time
	  if( ts > 0 )
		created = timestamp = ts;
	  if( deleted >= 0 )
		val.put( PeebleData.LocDistro._SYNC_DELETED, Long.toString( deleted ) );
	  if( created >= 0 )
		val.put( PeebleData.LocDistro._SYNC_CREATED, Long.toString( created ) );
	  if( timestamp >= 0 )
		val.put( PeebleData.LocDistro._SYNC_VERSION, Long.toString( timestamp ) );
	  cr.insert( PeebleData.LocDistro.CONTENT_URI,val );
	  resultValue = true;
	} else {
// else update the counter and sync timestamps
	  int counterColumnIndex = c.getColumnIndex( PeebleData.LocDistro.COUNTER );
	  int syncVersionColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_VERSION );
	  int syncDeletedColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_DELETED );
	  int syncCreatedColumnIndex = c.getColumnIndex( PeebleData.LocDistro._SYNC_CREATED );
	  c.updateString( counterColumnIndex, sa[2] );
// if autostamped, set the "timestamp" field to the current time
	  if( ts > 0 )
		timestamp = ts;
	  if( deleted >= 0 )
		c.updateLong( syncDeletedColumnIndex, deleted );
	  if( created >= 0 )
		c.updateLong( syncCreatedColumnIndex, created );
	  if( timestamp >= 0 )
		c.updateLong( syncVersionColumnIndex, timestamp );
	  c.commitUpdates();
	  resultValue = false;
	  c.close();
	}
	return resultValue;
  }

/**
  * Sets the data of the entry, the metainfo is set automatically depending on whether
  * the entry existed before or not. This is a convenience 
  * function for applications so that they don't have to deal with metadata.
  * @param key Key to create/update
  * @param type MIME type of the entry
  * @param data InputStream that can be used to read the data from
  * @return true if the entry was created, false if it was modified
  */
  public boolean setEntryData( String key, String type, InputStream data ) {
	HashMap metainf = new HashMap();
	metainf.put( "autostamp","1" );
	return setEntryData( key,metainf,data );
  }

/**
  * Deletes the entry data. The method also updates the meta information associated with
  * the key allowing to leave "tombstones" - meta information needed to delete entries during
  * synchronization
  * @param key The key to delete
  * @param metainf Meta information to set. If null, the meta information entry will also be deleted.
  */
  public void deleteEntryData( String key, Map metainf ) {
	Uri uri = Uri.withAppendedPath( PeebleData.LocDistro.CONTENT_URI, "delete" );
	uri = Uri.withAppendedPath( uri,key );
// if metainf is null, we really delete the entry
    if( metainf == null )
	  cr.delete( uri, null, null );
	else
	  setEntryMetaInfo( key,metainf );
  }

/**
  * Deletes the entry, manipulates the metadata automatically
  * @param key the key to delete
  */
  public void deleteEntryData( String key ) {
    HashMap metainfo = new HashMap();
	metainfo.put( "deleted", new Long( System.currentTimeMillis() ) );
	deleteEntryData( key, metainfo );
  }

/**
  * Locks the database for synchronization operation. The database can be manipulated by presenting
  * the session key passed as lock argument.
  * @param sessionKey session key to lock the database for.
  * @return true if the locking was succesful, false if other sync process is locking the database.
  */
  public boolean lock( String sessionKey ) {
	if( lockKey != null )
		return false;
	lockKey = sessionKey;
	return true;
  }

/**
  * Commits the changes done during this synchronization session.
  * @param sessionKey Session key that was passed to lock()
  */
  public void commit( String sessionKey ) {
	if( !sessionKey.equals( lockKey ) )
		return;
	lockKey = null;
  }

}
