package hu.uw.pallergabor.peeble;

import android.app.Service;
import android.content.Intent;
import android.content.Context;
import android.content.ContentResolver;
import android.content.ContentValues;
import android.database.Cursor;
import android.database.ContentObserver;
import android.location.LocationManager;
import android.location.LocationProvider;
import android.location.Location;
import android.net.Uri; 
import android.os.Bundle;
import android.os.IBinder;
import android.os.Handler;
import android.os.DeadObjectException;
import android.util.Log;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Semaphore;

import mwfw.ISendingProtocolAdapter;
import mwfw.common.HttpProtocolAdapter;
import mwfw.sync.TimeStampSyncEngine;
import mwfw.sync.ISyncProtocolAdapter;
import mwfw.sync.IStorage;
import mwfw.sync.ISyncListener;
import mwfw.syncmlds.SyncClient;

/**
  * This service maintains the location distribution data tables. Location
  * update events end up in this service which then talks to PeebleDataProvider
  * to update the persistent structures.
  */
public class LocationScanningService extends Service {
  public static final int SYNC_SUCCESS = 0;
  public static final int SYNC_INPROGRESS = 1;
  public static final int SYNC_NOCONF = 2;

  public static final String DISPLAY_HIGHLIGHTS_ACTION = "hu.uw.pallergabor.peeble.intent.DISPLAY_HIGHLIGHTS_ACTION";
  private static final String LOCUPDATE_ACTION = "hu.uw.pallergabor.peeble.action.LOCATION_UPDATE";
  private static final String LOG_TAG = "Peeble_LocationScanningService";

  private static final String SYNCDB_LOCDISTRO_NAME = "locdistro";
  private static final String SYNCDB_TERMS_NAME = "terms";
  private static final String SYNCDB_LOCHIGHLIGHT_NAME = "lochighlight";

  private Handler serviceHandler = new Handler();
  private static LocationScanningService serviceRef = null;
  private Location prevLocation = null;
  private ConfigBlock currentConfiguration = null;
  private SettingsContentObserver settingsObserver = null; 
  private double lastLongitude = 0.0;
  private double lastLatitude = 0.0;
  private IHighlightActivityCallback highlightActivityCallback = null;
  private String lastHighlightIDs = null;
  private int debugState = 0;
  private boolean syncInProgress = false;
  private Semaphore syncSemaphore = new Semaphore( 0 );

/**
  * The ILocScanService is defined through IDL
  */
  private final ILocScanService.Stub binder = 
			new ILocScanService.Stub() {
    public Map getLastFix() {
	  HashMap m = new HashMap();
	  synchronized( this ) {
		m.put( "longitude", Double.toString( lastLongitude ) );
		m.put( "latitude", Double.toString( lastLatitude ) );
	  }
	  return (Map)m;
    }

	public String getLastHighlightIDs() {
	  return lastHighlightIDs;
	}

    public void setHighlightCallback( IBinder binder ) {
		Log.d( LOG_TAG,"setHighlightCallback" );		
		highlightActivityCallback = IHighlightActivityCallback.Stub.asInterface( binder );
    }

	public void removeHighlightCallback() {
		Log.d( LOG_TAG,"setHighlightCallback" );
		highlightActivityCallback = null;
	}

/**
  * Starts synchronization session.
  * @return 
  *  <ul>
  *   <li>SYNC_SUCCESS, if synchronization session was started correctly
  *   <li>SYNC_INPROGRESS if there was already a sync session running.
  *   <li>SYNC_NOCONF if the sync configuration parameters are not correctly set
  *  </ul>
  */
	public int startSync() {
	  Log.d( LOG_TAG, "***************** Synchronization started ****************" );
	  if( syncInProgress ) {
		Log.e( LOG_TAG, "*** Synchronization: already in progress" );
		return SYNC_INPROGRESS;
	  }
	  if( currentConfiguration.syncServerAddress == null ) {
		Log.e( LOG_TAG, "*** Synchronization: not configured" );
		return SYNC_NOCONF;
	  }
	  syncInProgress = true;
	  Thread t = new Thread( new SyncThread() );
	  t.start();
	  return SYNC_SUCCESS;
	}

	class SyncListener implements ISyncListener {
	  public void sessionStarted( String sessionKey,boolean client ) {
		Log.d( LOG_TAG, "Sync session started: "+sessionKey );
	  }

	  public void sessionEnded( String sessionKey,boolean client ) {
		Log.d( LOG_TAG, "Sync session ended: "+sessionKey );
		syncSemaphore.release();
	  }
	}

	class SyncThread implements Runnable {
	  public void run() {
	    HttpProtocolAdapter pa = new HttpProtocolAdapter();
	    SyncClient sc = new SyncClient();
	    TimeStampSyncEngine ts = new TimeStampSyncEngine();
	    LocDistroSyncStorageAdapter ssa = 
			new LocDistroSyncStorageAdapter( getContentResolver() );
	    LocHighlightSyncStorageAdapter ssh = 
			new LocHighlightSyncStorageAdapter( getContentResolver() );
	    TermsSyncStorageAdapter sst = 
			new TermsSyncStorageAdapter( getContentResolver() );
	    sc.setSendingProtocolAdapter( (ISendingProtocolAdapter)pa );
	    ts.setSyncProtocolAdapter( (ISyncProtocolAdapter)sc );
	    ts.setStorage( SYNCDB_LOCDISTRO_NAME, (IStorage)ssa );
	    ts.setStorage( SYNCDB_LOCHIGHLIGHT_NAME, (IStorage)ssh );
	    ts.setStorage( SYNCDB_TERMS_NAME, (IStorage)sst );
        SyncListener sl = new SyncListener();
	    ts.addSyncListener( (ISyncListener)sl );
	    pa.init();
	    sc.init();
	    ssa.init();
		ssh.init();
	    HashMap tsInit = new HashMap();
		String myIMEI = android.os.SystemProperties.get(android.telephony.TelephonyProperties.PROPERTY_IMEI); 
	    tsInit.put( "mwfw.sync.timestamp.source","imei:"+myIMEI );
	    ts.init( tsInit );
		try {
		  Log.d( LOG_TAG, "*** Synchronizing "+SYNCDB_LOCDISTRO_NAME );
		  ts.startSync( currentConfiguration.syncServerAddress,
					SYNCDB_LOCDISTRO_NAME );
		  syncSemaphore.acquire();
		  Log.d( LOG_TAG, "*** Synchronizing "+SYNCDB_TERMS_NAME );
		  ts.startSync( currentConfiguration.syncServerAddress,
					SYNCDB_TERMS_NAME );
		  syncSemaphore.acquire();
		  Log.d( LOG_TAG, " *** Synchronizing: waiting for the server ..."  );
		  try {
			Thread.sleep( 5*1000L );	// 5 sec
		  } catch( InterruptedException ex ) {}
		  Log.d( LOG_TAG, " *** Synchronizing "+SYNCDB_LOCHIGHLIGHT_NAME );
		  ts.startSync( currentConfiguration.syncServerAddress,
					SYNCDB_LOCHIGHLIGHT_NAME );
		  syncSemaphore.acquire();
		} catch( Exception ex ) {
		  Log.e( LOG_TAG, "Synchronization error", ex );
		}
		Log.d( LOG_TAG, "*** Synchronization: end sync" );
	    syncInProgress = false;
	  }
	}
  };


/* This internal class is used as feed to serviceHandler's message queue.
   Each instance of this class carries two location fixes (the current one
   and the previous one) and the message processor acts according to
   the received location (updates the location distribution tables and
   informs the user in case of location match) */
  class LocationMessageProcessor implements Runnable {
	Location location = null;
	Location prevLocation = null;

	public LocationMessageProcessor( Location location, Location prevLocation ) {
	  this.location = location;
	  this.prevLocation = prevLocation;
	}

	public void run() {
	  ContentResolver cr = getContentResolver();
// This is the time between the two location fixes. This determines the amount
// of "ticks" added to both location histogram column (previous and current)
	  long timeDiff = ( location.getTime() - prevLocation.getTime() ) / 1000L;
// If timeDiff is 60 seconds, the delta is 1.0.
	  double delta = (double)timeDiff / 60.0;
	  synchronized( binder ) {
	    lastLongitude = location.getLongitude();
		lastLatitude = location.getLatitude();
	  }
// Let's update the counters belonging to the last 
	  ContentValues val = new ContentValues();
	  String longitudeStr = Double.toString( location.getLongitude() );
	  val.put( PeebleData.LocDistro.LONGITUDE_CENTER, longitudeStr );
	  String latitudeStr = Double.toString( location.getLatitude() );
	  val.put( PeebleData.LocDistro.LATITUDE_CENTER, latitudeStr );
	  val.put( PeebleData.LocDistro.COUNTER,
				Double.toString( delta ) );
	  Uri resultURI = cr.insert( PeebleData.LocDistro.CONTENT_URI,val );
	  Log.d( LOG_TAG,"locationEvent (1): ("+location.getLongitude()+
						","+location.getLatitude()+") "+resultURI.toString() );
	  val.clear();
	  val.put( PeebleData.LocDistro.LONGITUDE_CENTER, 
				Double.toString( prevLocation.getLongitude() ) );
	  val.put( PeebleData.LocDistro.LATITUDE_CENTER, 
				Double.toString( prevLocation.getLatitude() ) );
	  val.put( PeebleData.LocDistro.COUNTER,
				Double.toString( delta ) );
	  resultURI = cr.insert( PeebleData.LocDistro.CONTENT_URI,val );
	  Log.d( LOG_TAG,"locationEvent (2): ("+prevLocation.getLongitude()+
						","+prevLocation.getLatitude()+") "+resultURI.toString() );
// Let's look for entries that are younger than display_timestamp+warningTimeout
	  if( currentConfiguration.warningEnabled ) {
	    long limitTimeStamp = System.currentTimeMillis() - 
					( currentConfiguration.warningTimeout*60*1000L ); 
	    Uri queryUri = 
		  Uri.withAppendedPath( PeebleData.LocHighlight.CONTENT_URI, "location" );
	    queryUri = Uri.withAppendedPath( queryUri,longitudeStr+","+latitudeStr );
	    String columns[] = { PeebleData.LocHighlight._ID };
	    StringBuffer buffer = new StringBuffer();
	    Cursor resultCursor = cr.query( 
						queryUri,
						columns,
						"display_timestamp < "+Long.toString( limitTimeStamp ),
						null,
						null );
	    int ctr = 0;
	    while( resultCursor.next() ) {
		  int idColumnIndex = 
			resultCursor.getColumnIndex( PeebleData.LocHighlight._ID );
		  int id = resultCursor.getInt( idColumnIndex );
		  if( ctr > 0 )
			buffer.append( "," );
		  buffer.append( Integer.toString( id ) );
		  ++ctr;
	    }
/*debug!
	    switch( debugState ) {
		  case 0:
			buffer = new StringBuffer( "1" );
			debugState = 1;
			break;

		  case 1:
			buffer = new StringBuffer( "2,3" );
			debugState = 2;
			break;

		  case 2:
			buffer = new StringBuffer( "" );
			debugState = 0;
			break;
	    }
	    ctr = buffer.length() == 0 ? 0 : 1;
//end debug */
	    lastHighlightIDs = new String( buffer );
	    Log.d( LOG_TAG, "launch parameters: "+lastHighlightIDs );
	    resultCursor.close();
	    if( ctr > 0 ) {
		  if( highlightActivityCallback != null ) {
		    try {
			  highlightActivityCallback.updateIDs( lastHighlightIDs );
		    } catch( DeadObjectException ex ) {}
		  } else {
	        Intent i = new Intent( DISPLAY_HIGHLIGHTS_ACTION );
		    i.addCategory( Intent.DEFAULT_CATEGORY );
	        i.addLaunchFlags( Intent.NEW_TASK_LAUNCH );
	        startActivity( i );
		  }
	    } else
	    if( highlightActivityCallback != null ) {
		  try {
		    highlightActivityCallback.remove();
		  } catch( DeadObjectException ex ) {}
		  highlightActivityCallback = null;
	    }
	  }
	}
  }

  class SettingsContentObserver extends ContentObserver {

	public SettingsContentObserver( Handler h ) {
	  super( h );
	}

	public void onChange(boolean selfChange) {
	  Log.d( LOG_TAG, "SettingsContentObserver.onChange( "+selfChange+")" );
	  reConfig();
	}
  }

/* This internal class is used to hold the configuration values read from the
 * content provider.
 */
  class ConfigBlock {
	long notificationInterval = 0L;		// location scanning interval in millisec
	int notificationDistance = 0;		// location scanning distance in meters
	boolean warningEnabled = false;		// enable highlight warnings
	int warningTimeout = 0;				// timeout of a proximity alert in minutes
	String locProviderName = null;		// name of the location provider
	String syncServerAddress = null;	// URL of the sync server
	String syncUser = null;				// Sync user name
	String syncPassword = null;			// Sync password
  }

/**
  * Returns reference to this service instance. The method is called by
  * classes from the same package. This is possible because the service
  * is a singleton.
  * @return Reference to this service instance
  */
  public static LocationScanningService getServiceRef() {
	return serviceRef;
  }

  public void locationEvent( Location l ) {
	Log.d( LOG_TAG, "longitude: "+l.getLongitude()+
					",latitude: "+l.getLatitude() );
	if( prevLocation == null ) 
	  prevLocation = l;
	else {
// We process the location message in the serviceHandler's context
// to get rid of the event receivers' restrictions.
	  LocationMessageProcessor mp = 
		new LocationMessageProcessor( l,prevLocation );
	  serviceHandler.post( mp );
	  prevLocation = l;
	}
  }

  @Override
  protected void onCreate() {
	super.onCreate();
	Log.d( LOG_TAG, "onCreate" );
	ContentResolver cr = getContentResolver();
	settingsObserver = new SettingsContentObserver( serviceHandler );
	cr.registerContentObserver( PeebleData.Settings.CONTENT_URI, true,
							settingsObserver );
	if( currentConfiguration == null ) {
      currentConfiguration = new ConfigBlock();
	  getConfig( currentConfiguration );
	}
  }

  @Override
  protected void onDestroy() {
	super.onDestroy();
    Log.d( LOG_TAG, "onDestroy" );
	ContentResolver cr = getContentResolver();
	if( settingsObserver != null ) {		// just paranoia
	  cr.unregisterContentObserver( settingsObserver );
	  settingsObserver = null;
	}
	disableEvents();
  }

  @Override
  protected void onStart(int startId, Bundle arguments) {
	super.onStart( startId, arguments );
	serviceRef = this;
    Log.d( LOG_TAG, "onStart" );
	if( currentConfiguration == null ) {
      currentConfiguration = new ConfigBlock();
	  getConfig( currentConfiguration );
	}
	enableEvents();
  }

  @Override
  public IBinder onBind(Intent intent) {
	return binder;
  }

// Reads the configuration values from the database
  private void getConfig( ConfigBlock cb ) {
	ContentResolver cr = getContentResolver();
	String columns[] = { "name", "value" };
	Cursor resultCursor = cr.query( 
		PeebleData.Settings.CONTENT_URI, 
		columns,
		null,
		null,
		null );
	HashMap<String,String> m = new HashMap<String,String>();
	if( resultCursor != null ) {
	  while( resultCursor.next() ) {
	    int nameIndex = resultCursor.getColumnIndex( "name" );
	    int valueIndex = resultCursor.getColumnIndex( "value" );
	    String nameValue = resultCursor.getString( nameIndex );
	    String valueValue = resultCursor.getString( valueIndex );
	    m.put( nameValue, valueValue );
	  }
	  resultCursor.close();
    }

	cb.notificationInterval = Config.DEFAULTVALUE_NOTIFICATIONINTERVAL;
	String notificationIntervalStr = m.get( Config.CONFIGVALUE_NOTIFICATIONINTERVAL );
	if( notificationIntervalStr != null ) {
	  try {
		cb.notificationInterval = Long.parseLong( notificationIntervalStr ) * 1000L;
	  } catch( NumberFormatException ex ) {}
	}

	cb.notificationDistance = Config.DEFAULTVALUE_NOTIFICATIONDISTANCE;
	String notificationDistanceStr = m.get( Config.CONFIGVALUE_NOTIFICATIONDISTANCE );
	if( notificationDistanceStr != null ) {
	  try {
		cb.notificationDistance = Integer.parseInt( notificationDistanceStr );
	  } catch( NumberFormatException ex ) {}
	}

	String warningEnabledStr = m.get( Config.CONFIGVALUE_WARNINGENABLED );
	if( warningEnabledStr == null )
		warningEnabledStr = Config.DEFAULTVALUE_WARNINGENABLED;
	cb.warningEnabled = warningEnabledStr.equals( "true" );

	cb.warningTimeout = Config.DEFAULTVALUE_WARNINGTIMEOUT;
	String warningTimeoutStr = m.get( Config.CONFIGVALUE_WARNINGTIMEOUT );
	if( warningTimeoutStr == null ) {
	  try {
		cb.warningTimeout = Integer.parseInt( warningTimeoutStr );
	  } catch( NumberFormatException ex ) {}
	}

	cb.locProviderName = m.get( Config.CONFIGVALUE_LOCPROVIDER );
	cb.syncServerAddress = m.get( Config.CONFIGVALUE_SYNCADDRESS );
	cb.syncUser = m.get( Config.CONFIGVALUE_SYNCUSER );
	cb.syncPassword = m.get( Config.CONFIGVALUE_SYNCPASSWORD );
  }

// Re-reads configuration and updates location scanning operation accordingly
  private void reConfig() {
	ConfigBlock cb = new ConfigBlock();
	getConfig( cb );
	if( 
		( ( currentConfiguration.locProviderName != null ) &&
		  ( !currentConfiguration.locProviderName.equals( cb.locProviderName ) ) ) 
		|| 
		( currentConfiguration.notificationInterval != cb.notificationInterval )
		||
		( currentConfiguration.notificationDistance != cb.notificationDistance ) )
		disableEvents();
	currentConfiguration = cb;
	if( currentConfiguration.locProviderName != null )
		enableEvents();
  }

  private void enableEvents() {
	if( currentConfiguration.locProviderName == null )
		Log.d( LOG_TAG,"locProviderName is not configured" );
	else {
      LocationManager lm = (LocationManager)getSystemService(Context.LOCATION_SERVICE);
	  LocationProvider locProvider = 
			(LocationProvider)lm.getProvider( currentConfiguration.locProviderName );	  
	  if( locProvider == null )
		Log.e( LOG_TAG, "Location provider name "+
					currentConfiguration.locProviderName+
					" not found" );
	  else {
		Intent fireIntent = new Intent( LOCUPDATE_ACTION );
		Log.d( LOG_TAG, "Location events enabled, provider: "+
						currentConfiguration.locProviderName+
						", notificationInterval: "+
						currentConfiguration.notificationInterval+
						", notificationDistance: "+
						currentConfiguration.notificationDistance );
		lm.requestUpdates( locProvider,
					currentConfiguration.notificationInterval,
					currentConfiguration.notificationDistance,
					fireIntent );
	  }
	}
  }

  private void disableEvents() {
    LocationManager lm = (LocationManager)getSystemService(Context.LOCATION_SERVICE);
	Intent fireIntent = new Intent( LOCUPDATE_ACTION );
	lm.removeUpdates( fireIntent );
  }
}
