package mwfw.sync;

import java.util.Map;
import java.util.Iterator;
import java.io.InputStream;

/**
  * Components implementing this interface provide persistent storage service for 
  * synchronization engines. The component must expose mwfw.sync.dbid service property.
  * The value of this property is the URI under which the persistent storage is accessible (e.g. ./contacts)
  */
public interface IStorage {

/**
  * Gets synchronization metainfo associated with the database as collection of key-value pairs.
  * The definition of these key-value pairs depend on the synchronization engine. 
  * @return the synchronization metainformation associated with the database
  */
  public Map getDatabaseMetaInfo() throws Exception;

/**
  * Sets synchronization metainfo associated with the database as collection of key-value pairs.
  * The definition of these key-value pairs depend on the synchronization engine. 
  * @param metainf 
  */
  public void setDatabaseMetaInfo( Map metainf ) throws Exception;


/**
  * Selects database keys according to certain criteria. The criteria is specified as an
  * LDAP query executed on metainfo keys and values
  * @param query map. Each key is a condition element. If it the map is null, all keys in the database will be selected.
  * @return Iterator that can be used to enumerate the keys satisfying the criteria
  */
  public Iterator selectKeys( Map queryMap ) throws Exception;
  

/**
  * Gets the meta info map associated with a key
  * @param key Key in the database to query
  * @return Meta information in the form of key-value pairs or null if there's no meta information 
  *         associated with the entry.
  */
  public Map getEntryMetaInfo( String key ) throws Exception;

/**
  * Gets the meta info map associated with a key
  * @param key Key in the database to query
  * @param metainf Meta info to associate with the key
  */
  public void setEntryMetaInfo( String key, Map metainf ) throws Exception;

/**
  * Gets the data of the entry
  * @param key Key to read
  * @return The entry data as InputStream. The entry data can be read from this stream. Returns
  *         null if the entry doesn't exist.
  */
  public InputStream getEntryData( String key ) throws Exception;

/**
  * Sets the data and the metainfo of the entry
  * @param key Key to update
  * @param metainf The meta information to be associated with the entry
  * @param data InputStream that can be used to read the data from
  * @return true if the entry was created, false if it was modified
  */
  public boolean setEntryData( String key, Map metainf, InputStream data ) throws Exception;

/**
  * Sets the data of the entry, the metainfo is set automatically depending on whether
  * the entry existed before or not. This is a convenience 
  * function for applications so that they don't have to deal with metadata.
  * @param key Key to create/update
  * @param type MIME type of the entry
  * @param data InputStream that can be used to read the data from
  * @return true if the entry was created, false if it was modified
  */
  public boolean setEntryData( String key, String type, InputStream data ) throws Exception;

/**
  * Deletes the entry data. The method also updates the meta information associated with
  * the key allowing to leave "tombstones" - meta information needed to delete entries during
  * synchronization
  * @param key The key to delete
  * @param metainf Meta information to set. If null, the meta information entry will also be deleted.
  */
  public void deleteEntryData( String key, Map metainf ) throws Exception;

/**
  * Deletes the entry, manipulates the metadata automatically
  * @param key the key to delete
  */
  public void deleteEntryData( String key ) throws Exception;

/**
  * Locks the database for synchronization operation. The database can be manipulated by presenting
  * the session key passed as lock argument.
  * @param sessionKey session key to lock the database for.
  * @return true if the locking was succesful, false if other sync process is locking the database.
  */
  public boolean lock( String sessionKey ) throws Exception;

/**
  * Commits the changes done during this synchronization session.
  * @param sessionKey Session key that was passed to lock()
  */
  public void commit( String sessionKey ) throws Exception;
}
