package mwfw.sync;

import java.util.Map;
import java.util.List;
import java.io.InputStream;

/**
  * This interface is the engine-side counterpart of the ISyncProtocolAdapter interface.
  * The sync protocol adapter and the sync engine are connected by a bidirectional interface -
  * after connection, both can call the other. In order to avoid circular service references,
  * the link from the sync protocol adapter to the sync engine is implemented as a callback.
  */
public interface ISyncProtocolAdapterCallBack {

/** 
  * Conveys the partner's metainformation like sync session timestamps, etc. Each Sync Protocol Adapter
  * defines its own set of metainformation,  
  * @param sessionKey Session key of the synchronization session 
  * @param metainf Meta information sent when the session is opened
  * @param client True, if this session was initiated by the local sync client, false
  *        if this session is initiated by an incoming sync session initiated by remote sync client.
  */
  public void syncMetaInfo( String sessionKey, Map metainf, boolean client ) throws Exception;
  
/**
  * Conveys a command querying the local replica of the database.
  * @param sessionKey Session key of the synchronization session
  * @param metainf Meta information associated with the query (may be null)
  * @param key Key of the entry to query
  * @return InputStream that can be used to read the entry or null if the entry is unknown
  */
  public InputStream syncGet( String sessionKey, Map metainf, String key ) throws Exception;

/**
  * Conveys a command returning value of a query operation
  * @param sessionKey Session key of the synchronization session
  * @param metainf Meta information associated with the query.
  * <ul>
  * <li>type: MIME type of the data in the results message
  * </ul>
  * @param handle Handle of the query command (see ISyncProtocolAdapter::sendQuery) that this results is associated with
  * @param key Key of the entry in the result message
  * @param InputStream that can be used to read the result
  */ 
  public void syncQueryResult( String sessionKey, int handle, Map metainf, String key, InputStream is ) throws Exception;  

/**
  * Requests the sync engine to send its commands to the server.
  * @param sessionKey Session key
  * @param adapt Sync protocol adapter to use for sending the commands
  */
  public void sendCommands( String sessionKey,ISyncProtocolAdapter adapt ) throws Exception; 

/**
  * Requests the sync engine to execute and add/replace operation
  * @param sessionKey Session Key of the synchronization session
  * @param replace true if this is a replace operation, false if this is an addition operation
  * @param metainf Meta information associated with the query.
  * <ul>
  * <li>type: MIME type of the data in the results message
  * </ul>
  * @param source_db URI of the database that originally contains the data
  * @param target_db URI of the database that receives the data
  * @param key ID of the data in the source database
  * @param InputStream that can be used to read the data by the sync engine
  */
  public void syncAddReplace( String sessionKey, boolean replace, Map metainf, 
  							String source_db, String target_db, String key, InputStream is ) throws Exception;


/**
  * Requests the sync engine to delete an item
  * @param sessionKey Session key of the synchronization session
  * @param target_db URI of the database where the item should be deleted
  * @param key Key of the item to be deleted
  */
  public void syncDelete( String sessionKey, String target_db, String key ) throws Exception;

/**
  * Informs the sync engine that this session ends
  * @param sessionKey Key of the session that ends
  */
  public void endSession( String sessionKey ) throws Exception;
}
