package mwfw.syncmlds.impl;

import org.xml.sax.ContentHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import java.util.ArrayList;
import java.io.OutputStream;

/**
  * Object representing the parsing process of a document. This is the SAX document handler
  * and stores in its member variables the information extracted from the document.
  */
public class ParserState extends DefaultHandler implements ContentHandler {
  public final static boolean debug = false;
  
  final static int START = 0;
  final static int END = 1;
  final static int CHRS = 2;
  
  final static int NONE = 0;
  final static int SYNCML = 1;
  final static int SYNCHDR = 2;
  final static int SYNCBODY = 3;
  
  public final static int AUTH_NONE = 0;
  public final static int AUTH_BASIC = 1;
  
  int mainstate;
  
  public int sessionID = -1;
  public int msgID = -1;
  public int maxMsgSize = -1;
  public String source = null;
  public String target = null;
  public int authType = -1;
  public String userID = null;
  public String password = null;
  public String sessionKey = null;
  public OutputStream rspStream = null;
  public SyncSession session = null;
  
  private IReceiveCommand command = null;
  private boolean request = false;
  
/**
  * Public constructor
  * @param command Reference of the object with notification interface
  * @param request True if a request message is to be parsed, false otherwise
  */
  public ParserState( IReceiveCommand command, boolean request ) {
    mainstate = NONE;
    this.command = command;
    this.request = request;
  }

  public void resetParser() {
    sessionID = -1;
    msgID = -1;
    maxMsgSize = -1;
    source = null;
    target = null;
    authType = -1;
    userID = null;
    password = null;
    sessionKey = null;
    session = null;
  }

/**
  * Utility: calculates the session key
  */
  public static String getSessionKey( String source, int session ) {
    return source + ":" + Integer.toString( session );
  }

  public void startElement( String namespace, String localname, String type, Attributes attributes )
				 throws SAXException {
	if( debug )
		System.out.println( "startElement: "+type );
	if( mainstate == SYNCBODY ) {
	  if( msgID < 1 || msgID > 3 )
	  	throw new SAXException( "Illegal MsgID: "+msgID );
	  if( request ) {
	    switch( msgID ) {
	      case 1:
	    	processBodyPkg1Req( START, namespace, type );
	    	break;

	      case 2:
	      	processBodyPkg2ReqRsp( START, namespace, type );
	      	break;
	      	
		  case 3:
		    initBodyVariablesPkg1Rsp();
		    break;
		  
	    }
	  } else {
	    switch( msgID ) {
	      case 1:
	      case 3:
	    	processBodyPkg1Rsp( START, namespace, type );
	    	break;

	      case 2:
	      	processBodyPkg2ReqRsp( START, namespace, type );
	      	break;
	    }
	  }
	} else
	if( type.equals( "SyncML" ) ) {
	  if( mainstate != NONE ) 
	  	throw new SAXException( "<SyncML> tag must be the outermost tag" );
	  mainstate = SYNCML;
	} else
	if( mainstate == SYNCHDR )
		processHdr( START,namespace,type );
	else 
	if( type.equals( "SyncHdr" ) ) {
	  if( mainstate != SYNCML ) 
	  	throw new SAXException( "<SyncHdr> must be child of <SyncML>" );
	  mainstate = SYNCHDR;
	  hdrstate = SYNCHDR;
	  initHeaderVariables();
	} else
	if( type.equals( "SyncBody" ) ) {
	  if( mainstate != SYNCML ) 
	  	throw new SAXException( "<SyncBody> must be child of <SyncML>" );
	  mainstate = SYNCBODY;
	  bodystate = SYNCBODY;
	  if( msgID < 1 || msgID > 3 )
	  	throw new SAXException( "Illegal MsgID: "+msgID );
	  if( request ) {
	    switch( msgID ) {
	      case 1:
		    initBodyVariablesPkg1Req();
		    break;
		  
		  case 2:
		  	initBodyVariablesPkg2ReqRsp();
		  	break;
		  	
		  case 3:
		    initBodyVariablesPkg1Rsp();
		    break;
	    }
	  } else {
	    switch( msgID ) {
	      case 1:
	      case 3:
		    initBodyVariablesPkg1Rsp();
		    break;
		 
		  case 2:
		  	initBodyVariablesPkg2ReqRsp();
		  	break;
		}
	  }
	} else
		throw new SAXException( "unknown tag in SyncML context: "+namespace+"#"+type );
  }

  public void endElement( String namespace, String localname, String type ) throws SAXException {
    if( debug )
		System.out.println( "endElement: "+type );
    if( type.equals( "SyncHdr" ) ) {
      if( sessionID < 0 )
      		  	throw new SAXException( "no <SessionID> in SyncHdr!" );
      if( msgID < 0 )
      		  	throw new SAXException( "no <MsgID> in SyncHdr!" );
      if( source == null )
      		  	throw new SAXException( "no <Source> in SyncHdr!" );
      if( target == null )
      		  	throw new SAXException( "no <Target> in SyncHdr!" );
      sessionKey = getSessionKey( source, sessionID );
      mainstate = SYNCML;
      if( ( authType != AUTH_NONE ) &&
      	( ( userID == null ) || ( password == null ) ) )
      		      throw new SAXException( "Authentication credentials have not been supplied" );
      if( command != null )
      	command.endSyncHdr( this );
    } else
    if( mainstate == SYNCHDR )
    	processHdr( END,namespace,type );
    else
    if( type.equals( "SyncBody" ) ) {
      mainstate = SYNCML;
    } else
    if( mainstate == SYNCBODY ) {
	  if( msgID < 1 || msgID > 3 )
	  	throw new SAXException( "Illegal MsgID: "+msgID );
      if( request ) {
        switch( msgID ) {
          case 1:
        	processBodyPkg1Req( END,namespace,type );
        	break;

	      case 2:
	      	processBodyPkg2ReqRsp( END, namespace, type );
	      	break;
        	
          case 3:
        	processBodyPkg1Rsp( END,namespace,type );
        	break;          
        }
      } else {
        switch( msgID ) {
          case 1:
          case 3:
        	processBodyPkg1Rsp( END,namespace,type );
        	break;

	      case 2:
	      	processBodyPkg2ReqRsp( END, namespace, type );
	      	break;
	    }
      }
    } else
    if( type.equals( "SyncML" ) ) {
      if( command != null )
      	command.endMessage( sessionKey, msgID );
      mainstate = NONE;
    } 
  }
  
  public void characters( char[] ch, int start, int len ) throws SAXException {
    String text = new String( ch, start, len );
    String text1 = text.trim();
    if( text1.length() == 0 )
    	return;
    if( debug )
    	System.out.println( "characters: "+text1 );
    if( mainstate == SYNCHDR )
      processHdr( CHRS,null,text1 );
    else 
    if( mainstate == SYNCBODY ) {
	  if( msgID < 1 || msgID > 3 )
	  	throw new SAXException( "Illegal MsgID: "+msgID );
      if( request ) {
        switch( msgID ) {
          case 1:
        	processBodyPkg1Req( CHRS,null,text1 );
        	break;

	      case 2:
	      	processBodyPkg2ReqRsp( CHRS, null, text1 );
	      	break;

          case 3:
        	processBodyPkg1Rsp( CHRS,null,text1 );
        	break;
        }
      } else {
        switch( msgID ) {
          case 1:
          case 3:
        	processBodyPkg1Rsp( CHRS,null,text1 );
        	break;

	      case 2:
	      	processBodyPkg2ReqRsp( CHRS, null, text1 );
	      	break;
        }
      }
    }
    	
  }
  
// ------------------------------ private methods  

// ------------------------- Meta
  final static int META_BODY = 100;
  final static int META_MAXMSGSIZE = 101;
  final static int META_TYPE = 102;
  final static int META_ANCHOR = 103;
  final static int META_LAST = 104;
  final static int META_NEXT = 105;
  
  private int metaState;
  private String meta_Type;
  private int meta_maxMsgSize;
  private long meta_Last;
  private long meta_Next;
  private String meta_LastString;
  private String meta_NextString;
  private String meta_maxMsgSizeString;
 
  private void initMetaVariables() {
    metaState = META_BODY;
    meta_maxMsgSize = -1;
    meta_Type = null;
    meta_Last = -1;
    meta_Next = -1;
	meta_LastString = null;
	meta_NextString = null;
	meta_maxMsgSizeString = null;
  }
  
  private void processMeta( int context, String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
            if( parm.equals( "MaxMsgSize" ) ) {
      		  if( metaState != META_BODY )
      		  	throw new SAXException( "<MaxMsgSize> must be child of <Meta>" );
      		  metaState = META_MAXMSGSIZE;
            } else
            if( parm.equals( "Type" ) ) {
              if( metaState != META_BODY )
      		  	throw new SAXException( "<Type> must be child of <Meta>" );
      		  metaState = META_TYPE;
            } else 
            if( parm.equals( "Anchor" ) ) {
              if( metaState != META_BODY )
      		  	throw new SAXException( "<Anchor> must be child of <Meta>" );
      		  metaState = META_ANCHOR;
            } else
			if(  parm.equals( "Last" ) ) {
              if( metaState != META_ANCHOR )
      		  	throw new SAXException( "<Last> must be child of <Anchor>" );
      		  metaState = META_LAST;
			} else      
			if(  parm.equals( "Next" ) ) {
              if( metaState != META_ANCHOR )
      		  	throw new SAXException( "<Next> must be child of <Anchor>" );
      		  metaState = META_NEXT;
			} else      
            	throw new SAXException( "unknown tag in Meta context: "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "MaxMsgSize" ) ) {
      		  try {
      		    meta_maxMsgSize = Integer.parseInt( parm );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "Maximum message size "+parm+" is not a number" );
      		  }
      		  metaState = META_BODY;
			} else
			if( parm.equals( "Type" ) ||
      		    	parm.equals( "Anchor" ) ) {
      		  metaState = META_BODY;
      		} else
      		if( parm.equals( "Last" ) ) {
      		  try {
      		    meta_Last = Long.parseLong( meta_LastString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "Last anchor "+meta_LastString+" is not a number" );
      		  }
      		  metaState = META_ANCHOR;
			} else
			if( parm.equals( "Next" ) ) {
      		  try {
      		    meta_Next = Long.parseLong( meta_NextString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "Next anchor "+meta_NextString+" is not a number" );
      		  }
      		  metaState = META_ANCHOR;
      		}
      		break;
      		
      case CHRS:
      		if( metaState == META_MAXMSGSIZE ) {
			  if( meta_maxMsgSizeString == null )
					meta_maxMsgSizeString = parm;
			  else
					meta_maxMsgSizeString += parm;
      		} else
      		if( metaState == META_TYPE ) {
			  if( meta_Type == null )
					meta_Type = parm;
			  else
					meta_Type += parm;
      		} else
      		if( metaState == META_LAST ) {
			  if( meta_LastString == null )
					meta_LastString = parm;
			  else
					meta_LastString += parm;
      		} else
      		if( metaState == META_NEXT ) {
			  if( meta_NextString == null )
					meta_NextString = parm;
			  else
					meta_NextString += parm;
      		}
      		break;
    }
  } 

// ----------------------------------------- Data
  private StringBuffer data_Buffer = null;
  
  private void initDataVariables() {
    data_Buffer = new StringBuffer();
  }

  private void processData( int context, String namespace, String parm ) {
    switch( context ) {
      case START:
      	data_Buffer.append( "<"+parm );
      	if( namespace != null )
      		data_Buffer.append( " xmlns='"+namespace+"'" );
      	data_Buffer.append( ">" );
      	break;

	  case END:
	  	data_Buffer.append( "</"+parm+">" );
	  	break;
	  	
	  case CHRS:
	  	data_Buffer.append( parm );
	  	break;
	}
  }
  
// ----------------------------------------- Item
  final static int ITEM_BODY = 100;
  final static int ITEM_SOURCE = 101;
  final static int ITEM_TARGET = 102;
  final static int ITEM_META = 103;
  final static int ITEM_DATA = 104;
  final static int ITEM_LOCURI = 105;
  private int itemState;
  
  private int item_LocURIPrevState;
  private String item_Source;
  private String item_Target;
  private long item_Last;
  private long item_Next;
  private String item_Data;
  private String item_LocURI;
  
  private void initItemVariables() {
    itemState = ITEM_BODY;
    item_Source = null;
    item_Target = null;
    item_Last = -1;
    item_Next = -1;
    item_Data = null;
    item_LocURI = null;
  }

  private void processItem( int context, String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
      		if( itemState == ITEM_DATA )
      			processData( context,namespace,parm );
      		else
      		if( itemState == ITEM_META )
      			processMeta( context,namespace,parm );
      		else
      		if( parm.equals( "Data" ) ) {
      		  if( itemState != ITEM_BODY )
      		  	throw new SAXException( "<Data> must be child of <Item>" );
      		  itemState = ITEM_DATA;
      		  initDataVariables();
      		} else
      		if( parm.equals( "Meta" ) ) {
      		  if( itemState != ITEM_BODY )
      		  	throw new SAXException( "<Meta> must be child of <Item>" );
      		  itemState = ITEM_META;
      		  initMetaVariables();
      		} else
            if( parm.equals( "Target" ) ) {
      		  if( itemState != ITEM_BODY )
      		  	throw new SAXException( "<Target> must be child of <Item>" );
      		  item_LocURI = null;
      		  itemState = ITEM_TARGET;
            } else
            if( parm.equals( "Source" ) ) {
              if( itemState != ITEM_BODY )
      		  	throw new SAXException( "<Source> must be child of <Item>" );
      		  item_LocURI = null;
      		  itemState = ITEM_SOURCE;
            } else      
            if( parm.equals( "LocURI" ) ) {
              if( ( itemState != ITEM_SOURCE ) && ( itemState != ITEM_TARGET ) )
      		  	throw new SAXException( "<LocURI> must be child of <Source> or <Target>" );
      		  item_LocURIPrevState = itemState;
      		  itemState = ITEM_LOCURI;
            } else
            	throw new SAXException( "unknown tag in Item context: "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "Data" ) ) {
      		  item_Data = new String( data_Buffer );
      		  itemState = ITEM_BODY;
      		} else
      		if( itemState == ITEM_DATA )
      			processData( context,namespace,parm );
      		else
      		if( parm.equals( "Meta" ) ) {
      		  itemState = ITEM_BODY;
      		  if( meta_Last >= 0 )
      		  	item_Last = meta_Last;
      		  if( meta_Next >= 0 )
      		  	item_Next = meta_Next;
      		} else
      		if( itemState == ITEM_META )
      			processMeta( context,namespace,parm );
      		else
      		if( parm.equals( "Target" ) ) {
      		  itemState = ITEM_BODY;
      		  if( item_LocURI != null )
      		  	item_Target = item_LocURI;
      		} else
      		if( parm.equals( "Source" ) ) {
      		  itemState = ITEM_BODY;
      		  if( item_LocURI != null )
      		  	item_Source = item_LocURI;
      		} else
      		if( parm.equals( "LocURI" ) ) {
      		  itemState = item_LocURIPrevState;
      		}
      		break;
      		
      case CHRS:
      		if( itemState == ITEM_DATA )
      			processData( context,namespace,parm );
      		else
      		if( itemState == ITEM_META )
      			processMeta( context,namespace,parm );
      		else
      		if( itemState == ITEM_SOURCE ) {
			  if( item_Source == null )
				item_Source = parm;
			  else
      		  	item_Source += parm;
      		} else
      		if( itemState == ITEM_TARGET ) {
			  if( item_Target == null )
				item_Target = parm;
			  else
				item_Target += parm;
      		} else
      		if( itemState == ITEM_LOCURI ) {
			  if( item_LocURI == null )
				item_LocURI = parm;
			  else
				item_LocURI += parm;
      		}
      		break;
    }
  } 

// ----------------------------------------- SyncHdr
  private int hdrstate;
  final static int HDR_VERDTD = 3;
  final static int HDR_VERPROTO = 4;
  final static int HDR_SESSIONID = 5;
  final static int HDR_SOURCE = 6;
  final static int HDR_TARGET = 7;
  final static int HDR_MSGID = 8;
  final static int HDR_CRED = 9;
  final static int HDR_META = 10;
  final static int HDR_CRED_DATA = 12;
  private int metaPrevState;
  private String verDTD = null;
  private String verProto = null;
  
  private void initHeaderVariables() {
    sessionID = -1;
    msgID = -1;
    source = null;
    target = null;
    authType = AUTH_NONE;
    userID = null;
    password = null;
    maxMsgSize = -1;
	verDTD = null;
	verProto = null;
  }

  private void processHdr( int context, String namespace, String parm ) throws SAXException {
    if( debug )
	    System.out.println( "processHdr: "+context+" ->"+parm );
    switch( context ) {
      case START:
      		if( hdrstate == HDR_META )
      			processMeta( context,namespace,parm );
      		else
      		if( parm.equals( "VerDTD" ) ) {
      		  if( hdrstate != SYNCHDR )
      		  	throw new SAXException( "<VerDTD> must be child of <SyncHdr>" );
      		  hdrstate = HDR_VERDTD;
            } else
            if( parm.equals( "VerProto" ) ) {
      		  if( hdrstate != SYNCHDR )
      		  	throw new SAXException( "<VerProto> must be child of <SyncHdr>" );
      		  hdrstate = HDR_VERPROTO;
            } else
            if( parm.equals( "SessionID" ) ) {
      		  if( hdrstate != SYNCHDR )
      		  	throw new SAXException( "<SessionID> must be child of <SyncHdr>" );
      		  hdrstate = HDR_SESSIONID;              
            } else
            if( parm.equals( "MsgID" ) ) {
      		  if( hdrstate != SYNCHDR )
      		  	throw new SAXException( "<MsgID> must be child of <SyncHdr>" );
      		  hdrstate = HDR_MSGID;              
            } else
            if( parm.equals( "Source" ) ) {
      		  if( hdrstate != SYNCHDR )
      		  	throw new SAXException( "<Source> must be child of <SyncHdr>" );
      		  hdrstate = HDR_SOURCE;              
            } else
            if( parm.equals( "Target" ) ) {
      		  if( hdrstate != SYNCHDR )
      		  	throw new SAXException( "<Target> must be child of <SyncHdr>" );
      		  hdrstate = HDR_TARGET;              
            } else
            if( parm.equals( "Cred" ) ) {
      		  if( hdrstate != SYNCHDR )
      		  	throw new SAXException( "<Cred> must be child of <SyncHdr>" );
      		  hdrstate = HDR_CRED;
            } else
            if( parm.equals( "Meta" ) ) {
      		  if( ( hdrstate != HDR_CRED ) && ( hdrstate == SYNCHDR ) )
      		  	throw new SAXException( "<Meta> must be child of <SyncHdr> or <Cred>" );
			  metaPrevState = hdrstate;
			  initMetaVariables();
      		  hdrstate = HDR_META;              
            } else
            if( parm.equals( "Data" ) ) {
      		  if( hdrstate != HDR_CRED )
      		  	throw new SAXException( "<Data> must be child of <Cred>" );
      		  hdrstate = HDR_CRED_DATA;
            } else
            	throw new SAXException( "unknown tag in SyncHdr context: "+namespace+"#"+parm );
      		break;
      		
      case END:
            if( parm.equals( "Meta" ) ) {
      		  hdrstate = metaPrevState;
      		  if( meta_maxMsgSize >= 0 ) {
      		    if( hdrstate != SYNCHDR )
      		    	throw new SAXException( "<MaxMsgSize> must be child of <SyncHdr>" );
      		  	maxMsgSize = meta_maxMsgSize;
      		  }
      		  if( meta_Type != null ) {
      		    if( hdrstate != HDR_CRED )
      		    	throw new SAXException( "<Type> must be subelement of <Cred>" );
      		    if( meta_Type != "syncml:auth-basic" )
      		    	throw new SAXException( "Only syncml:auth-basic is supported ("+meta_Type+")" );
      		    authType = AUTH_BASIC;
      		  }
      		} else
      		if( hdrstate == HDR_META )
      			processMeta( context,namespace,parm );
      		else 
      		if( parm.equals( "VerDTD" ) ) {
      		  if( !"1.1".equals( verDTD ) )
      		  	throw new SAXException( "VerDTD value must be 1.1" );
      		  hdrstate = SYNCHDR;
      		} else
      		if( parm.equals( "VerProto" ) ) {
      		  if( !"SyncML/1.1".equals( verProto ) )
      		  	throw new SAXException( "VerProto value must be SyncML/1.1" );
      		  hdrstate = SYNCHDR;
      		} else
      		if( parm.equals( "SessionID" ) ) {
      		  hdrstate = SYNCHDR;
      		} else
      		if( parm.equals( "MsgID" ) ) {
      		  hdrstate = SYNCHDR;
      		} else
      		if( parm.equals( "Source" ) ) {
      		  hdrstate = SYNCHDR;
      		} else
      		if( parm.equals( "Target" ) ) {
      		  hdrstate = SYNCHDR;
      		} else
      		if( parm.equals( "Cred" ) ) {
      		  hdrstate = SYNCHDR;
      		} else
      		if( parm.equals( "Data" ) ) {
      		  hdrstate = HDR_CRED;
      		}
      		break;

      case CHRS:
      		if( hdrstate == HDR_META ) 
      		  	processMeta( context,namespace,parm );
      		else
      		if( hdrstate == HDR_VERDTD ) {
			  if( verDTD == null )
					verDTD = parm;
			  else
					verDTD += parm;
      		} else
      		if( hdrstate == HDR_VERPROTO ) {
			  if( verProto == null )
					verProto = parm;
			  else
					verProto += parm;
      		} else 
      		if( hdrstate == HDR_SESSIONID ) {
      		  try {
      		    sessionID = Integer.parseInt( parm );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "Session ID "+parm+" is not a number" );
      		  }
      		} else
      		if( hdrstate == HDR_MSGID ) {
      		  try {
      		    msgID = Integer.parseInt( parm );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "Message ID "+parm+" is not a number" );
      		  }
      		} else
      		if( hdrstate == HDR_SOURCE ) {
      		  if( parm.length() == 0 )
      		  	throw new SAXException( "<Source> must have a non-empty content" );
			  if( source == null )
				source = parm;
			  else
				source += parm;	
      		} else
       		if( hdrstate == HDR_TARGET ) {
      		  if( parm.length() == 0 )
      		  	throw new SAXException( "<Target> must have a non-empty content" );
			  if( target == null )
				target = parm;
			  else
				target += parm;	
      		} else 
      		if( hdrstate == HDR_CRED_DATA ) {
      		  byte cred[] = Base64.decode( parm );
      		  String creds = new String( cred,0 );
      		  int idx = creds.indexOf( ':' );
      		  if( ( idx < 0 ) || ( idx+1 == creds.length() ) )
      		  	throw new SAXException( "credentials must be in username:password format (received: "+creds+")" );
      		  userID = creds.substring( 0,idx );
      		  password = creds.substring( idx+1 );
      		}
      		break;
    }
  }

// ----------------------------- SyncBody for Pkg#1 Request
  private int bodystate;
  private int prevBodyState;
  private int syncCmdID;
  private String syncCmdIDString = null;
  private String syncSource;
  private String syncTarget;
  final static int BODY_ALERT = 50;
  final static int BODY_PUT = 51;
  final static int BODY_GET = 52;
  final static int BODY_STATUS = 53;
  final static int BODY_RESULTS = 54;
  final static int BODY_SYNC = 55;
  final static int BODY_ADD = 56;
  final static int BODY_REPLACE = 57;
  final static int BODY_DELETE = 58;
  final static int BODY_SYNC_CMDID = 59;
  final static int BODY_SYNC_SOURCE = 60;
  final static int BODY_SYNC_TARGET = 61;
  
  private void initBodyVariablesPkg1Req() {
    bodystate = SYNCBODY;
  }

  private void processBodyPkg1Req( int context, String namespace, String parm ) throws SAXException {
    if( debug )
    	System.out.println( "processBodyPkg1Req: "+context+" -> "+parm );
    switch( context ) {
      case START:
      		if( bodystate == BODY_ALERT )
      			processAlert( context,namespace,parm );
      		else
      		if( parm.equals( "Alert" ) ) {
      		  initAlertVariables();
      		  bodystate = BODY_ALERT;
      		} else
      		if( bodystate == BODY_PUT )
      			processPut( context,namespace,parm );
      		else
      		if( parm.equals( "Put" ) ) {
      		  initPutVariables();
      		  bodystate = BODY_PUT;
			} else      		
      		if( bodystate == BODY_GET )
      			processGet( context,namespace,parm );
      		else
      		if( parm.equals( "Get" ) ) {
      		  initGetVariables();
      		  bodystate = BODY_GET;
			} else
			if( !parm.equals( "Final" ) )
				throw new SAXException( "unknown tag in SyncBody(MsgID:1) context: "+namespace+"#"+parm );      		
      		break;
      		
      case END:
      		if( parm.equals( "Alert" ) ) {
      		  bodystate = SYNCBODY;
      		  if( alert_Data == null )
      		  	throw new SAXException( "<Alert> in Pkg#1: Data not present" );
      		  int type = -1;
      		  try {
      		    type = Integer.parseInt( alert_Data );
      		  } catch( NumberFormatException ex ) {
      		  	throw new SAXException( "<Alert> in Pkg#1: invalid sync type: "+type );
      		  }
      		  if( alert_Source == null )
      		  	throw new SAXException( "<Alert> in Pkg#1: Source not set" );
      		  if( alert_Target == null )
      		  	throw new SAXException( "<Alert> in Pkg#1: Target not set" );
      		  if( alert_Last < 0 )
      		  	throw new SAXException( "<Alert> in Pkg#1: Last not set" );
      		  if( alert_Next < 0 )
      		  	throw new SAXException( "<Alert> in Pkg#1: Next not set" );
      		  if( command != null )
      		  	command.dbAlertInPackage1( sessionKey, cmdID, type,
      		  			alert_Source, alert_Target, alert_Last, alert_Next );
      		} else
      		if( bodystate == BODY_ALERT )
      			processAlert( context,namespace,parm );
      		else
      		if( parm.equals( "Put" ) ) {
      		  bodystate = SYNCBODY;
      		  if( !put_Type.equals( "application/vnd.syncml-devinf+xml" ) ) {
				System.err.println( "put_Type: "+put_Type );
      		  	throw new SAXException( "Devinfo type must be application/vnd.syncml-devinf+xml in Put in Pkg#1" );
			  }
      		  if( !put_Source.equals( "./devinf11" ) )
      		  	throw new SAXException( "Devinfo source must be ./devinf11 in Put in Pkg#1" );
      		  if( put_Data == null )
      		  	throw new SAXException( "<Data> missing in Put in Pkg#1" );
      		  if( command != null )
      		  	command.putInPackage1( sessionKey,cmdID, put_Data );
      		} else
      		if( bodystate == BODY_PUT )
      		  	processPut( context,namespace,parm );
      		else
      		if( parm.equals( "Get" ) ) {
      		  bodystate = SYNCBODY;
      		  if( !get_Type.equals( "application/vnd.syncml-devinf+xml" ) ) {
				System.err.println( "get_Type: "+get_Type );
      		  	throw new SAXException( "Devinfo type must be application/vnd.syncml-devinf+xml in Get in Pkg#1" );
			  }
      		  if( !get_Target.equals( "./devinf11" ) )
      		  		throw new SAXException( "Devinfo target must be ./devinf11 in Put in Pkg#1" );
      		  if( command != null )
      		  	command.getCommand( sessionKey,msgID,cmdID,get_Type,get_Target );
      		} else
      		if( bodystate == BODY_GET )
      		  	processGet( context,namespace,parm );
      		break;
      		
      case CHRS:
      		if( bodystate == BODY_PUT )
      		  	processPut( context,namespace,parm );
      		else
      		if( bodystate == BODY_ALERT )
      			processAlert( context,namespace,parm );
      		else
      		if( bodystate == BODY_GET )
      		  	processGet( context,namespace,parm );
      		break;
    }
  }

// ----------------------------- SyncBody for Pkg#1 Response, Pkg#3 Request, Response
  private void initBodyVariablesPkg1Rsp() {
    bodystate = SYNCBODY;
  }

  private void processBodyPkg1Rsp( int context, String namespace, String parm ) throws SAXException {
    if( debug )
    	System.out.println( "processBodyPkg1Rsp: "+context+" -> "+parm );
    switch( context ) {
      case START:
      		if( bodystate == BODY_STATUS )
      			processStatus( context,namespace,parm );
      		else
      		if( bodystate == BODY_RESULTS )
      			processResults( context,namespace,parm );
      		else
      		if( bodystate == BODY_ALERT )
      			processAlert( context,namespace,parm );
      		else
      		if( parm.equals( "Status" ) ) {
      		  bodystate = BODY_STATUS;
      		  initStatusVariables();
      		} else
      		if( parm.equals( "Results" ) ) {
      		  bodystate = BODY_RESULTS;
      		  initResultsVariables();
      		} else
      		if( parm.equals( "Alert" ) ) {
      		  bodystate = BODY_ALERT;
      		  initAlertVariables();
      		} else
      		if( !parm.equals( "Final" ) )
      			throw new SAXException( "Illegal tag in <SyncBody> (Msg1, response): "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "Status" ) ) {
      		  if( command != null )
      		  	command.statusCommand( sessionKey, msgID, cmdID, status_Data, status_MsgRef,status_cmdRef, 
      		  							status_Cmd, status_SourceRef, status_TargetRef );
      		  bodystate = SYNCBODY;
      		} else
      	    if( bodystate == BODY_STATUS )
      	    	processStatus( context,namespace,parm );
      	    else
      		if( parm.equals( "Results" ) ) {
      		  if( command != null )
      		  	command.resultsCommand( sessionKey, msgID, cmdID, results_MsgRef, results_CmdRef, 
      		  							results_Source, results_Type, results_Data );
      		  bodystate = SYNCBODY;
      		} else
      	    if( bodystate == BODY_RESULTS )
      	    	processResults( context,namespace,parm );
      		if( parm.equals( "Alert" ) ) {
      		  bodystate = SYNCBODY;
      		} else
      	    if( bodystate == BODY_ALERT )
      	    	processAlert( context,namespace,parm );
      		break;
      		
      case CHRS:
      	    if( bodystate == BODY_STATUS )
      	    	processStatus( context,namespace,parm );
      	    else
      	    if( bodystate == BODY_RESULTS )
      	    	processResults( context,namespace,parm );
      	    else
      	    if( bodystate == BODY_ALERT )
      	    	processAlert( context,namespace,parm );
      		break;
    }
  }

// ----------------------------- SyncBody for Pkg#2 Request and Response   
  private void initBodyVariablesPkg2ReqRsp() {
    bodystate = SYNCBODY;
    syncCmdID = -1;
	syncCmdIDString = null;
    syncSource = null;
    syncTarget = null;
  }

  private void processBodyPkg2ReqRsp( int context, String namespace, String parm ) throws SAXException {
    if( debug )
    	System.out.println( "processBodyPkg2ReqRsp: "+context+" -> "+parm );
    switch( context ) {
      case START:
      		if( bodystate == BODY_RESULTS )
      			processResults( context,namespace,parm );
      		else
      		if( bodystate == BODY_STATUS )
      			processStatus( context,namespace,parm );
      		else
      		if( bodystate == BODY_GET )
      			processGet( context,namespace,parm );
      		else
      		if( ( bodystate == BODY_ADD ) || ( bodystate == BODY_REPLACE ) )
      			processAdd( context,namespace,parm );
      		else
      		if( bodystate == BODY_DELETE )
      			processDelete( context,namespace,parm );
      		else
      		if( parm.equals( "Delete" ) ) {
      		  if( bodystate != BODY_SYNC )
      		  	throw new SAXException( "<Delete> must be child of <Sync>" );
      		  initDeleteVariables();
      		  bodystate = BODY_DELETE;
      		} else
      		if( parm.equals( "Results" ) ) {
      		  if( bodystate != SYNCBODY )
      		  	throw new SAXException( "<Results> must be child of <SyncBody>" );
      		  if( request )
      		  	throw new SAXException( "<Results> must not appear in Msg#2 request" );
      		  initResultsVariables();
      		  bodystate = BODY_RESULTS;
      		} else
      		if( parm.equals( "Add" ) ||
      			parm.equals( "Replace" ) ) {
      		  if( bodystate != BODY_SYNC )
      		  	throw new SAXException( "<Add> must be child of <Sync>" );
      		  initAddVariables( parm );
      		  if( parm.equals( "Add" ) )
      		  	bodystate = BODY_ADD;
      		  else
      		  	bodystate = BODY_REPLACE;
      		} else
      		if( parm.equals( "Sync" ) ) {
      		  if( bodystate != SYNCBODY )
      		  	throw new SAXException( "<Sync> must be child of <SyncBody>" );
      		  bodystate = BODY_SYNC;
      		} else
      		if( parm.equals( "CmdID" ) ) {
      		  if( bodystate != BODY_SYNC )
      		  	throw new SAXException( "<CmdID> must be child of <Sync>" );
      		  bodystate = BODY_SYNC_CMDID;
      		} else
      		if( parm.equals( "Source" ) ) {
      		  if( bodystate != BODY_SYNC )
      		  	throw new SAXException( "<Source> must be child of <Sync>" );
      		  bodystate = BODY_SYNC_SOURCE;
      		} else
      		if( parm.equals( "Target" ) ) {
      		  if( bodystate != BODY_SYNC )
      		  	throw new SAXException( "<Target> must be child of <Sync>" );
      		  bodystate = BODY_SYNC_TARGET;
      		} else
      		if( parm.equals( "Get" ) ) {
      		  if( bodystate != SYNCBODY )
      		  	throw new SAXException( "<Get> must be child of <SyncBody>" );
      		  initGetVariables();
      		  bodystate = BODY_GET;
      		} else
      		if( parm.equals( "Status" ) ) {
      		  if( bodystate != SYNCBODY )
      		  	throw new SAXException( "<Status> must be child of <SyncBody>" );
      		  initStatusVariables();
      		  bodystate = BODY_STATUS;
      		} else
      		if( !parm.equals( "Final" ) ) {
      		  String r = request ? "request" : " response";
      	      throw new SAXException( "Illegal tag in <SyncBody> (Msg2, "+r+"): "+namespace+"#"+parm );
      	    }
			break;
			
	  case END:
	        if( parm.equals( "Delete" ) ) {
	          if( command != null )
	          	command.deleteCommand( sessionKey, msgID, cmdID, syncTarget, delete_Target );
	          bodystate = BODY_SYNC;
	        } else
	        if( bodystate == BODY_DELETE )
	        	processDelete( context,namespace,parm );
	        else
	  		if( parm.equals( "Add" ) ||
	  			parm.equals( "Replace" ) ) {
      		  if( add_Source == null )
      		    	throw new SAXException( "Add/Replace: source not set" );
      		  if( add_Type == null )
      		    	throw new SAXException( "Add/Replace: Type not set" );
      		  if( add_Data == null )
      		    	throw new SAXException( "Add/Replace: no data" );
      		  if( command != null )
      		  	command.addReplaceCommand( sessionKey, bodystate == BODY_REPLACE, 
      		  							  msgID, cmdID, 
  										  syncSource, syncTarget, add_Source, add_Type, 
      		  							  add_Data );
	  		  bodystate = BODY_SYNC;	  		  
	  		} else
	  		if( ( bodystate == BODY_ADD ) || ( bodystate == BODY_REPLACE ) )
	  			processAdd( context,namespace,parm );
	  		else
	  		if( parm.equals( "Get" ) ) {
      		  if( command != null )
      		  	command.getCommand( sessionKey,msgID,cmdID,get_Type,get_Target );
	  		  bodystate = SYNCBODY;
	  		} else
	  		if( bodystate == BODY_GET )
	  			processGet( context,namespace,parm );
	  		else
	  		if( parm.equals( "Results" ) ) {
      		  if( command != null )
				command.resultsCommand( sessionKey, msgID, cmdID, results_MsgRef, results_CmdRef, 
      		  							results_Source, results_Type, results_Data );
      		  bodystate = SYNCBODY;
	  		} else
	  		if( bodystate == BODY_RESULTS )
	  			processResults( context,namespace,parm );
	  		else
	  		if( parm.equals( "Status" ) ) {
      		  if( command != null )
      		  	command.statusCommand( sessionKey, msgID, cmdID, status_Data, status_MsgRef,status_cmdRef, 
      		  							status_Cmd, status_SourceRef, status_TargetRef );
	  		  bodystate = SYNCBODY;
	  		} else
	  		if( bodystate == BODY_STATUS )
	  			processStatus( context,namespace,parm );
	  		else
	  		if( parm.equals( "Sync" ) ) {
	  		  bodystate = SYNCBODY;
	  		  if( syncCmdID < 0 )
	  		  	throw new SAXException( "CmdID in Sync not set" );
	  		  if( syncSource == null )
	  		  	throw new SAXException( "Source in Sync not set" );
	  		  if( syncTarget == null )
	  		  	throw new SAXException( "Target in Sync not set" );
      		  if( command != null )
      		  	command.syncCommand( sessionKey, msgID, cmdID, syncSource, syncTarget );
	  		} else
	  		if( parm.equals( "CmdID" ) ) {
      		  try {
      		    syncCmdID = Integer.parseInt( syncCmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+syncCmdIDString+" is not a number" );
      		  }
	  		  if( syncCmdID < 0 )
	  		  	throw new SAXException( "CmdID in Sync has no content" );
	  		  bodystate = BODY_SYNC;
	  		} else
	  		if( parm.equals( "Source" ) ) {
	  		  if( syncSource == null )
	  		  	throw new SAXException( "Source in Sync has no content" );
	  		  bodystate = BODY_SYNC;
	  		} else
	  		if( parm.equals( "Target" ) ) {
	  		  if( syncTarget == null )
	  		  	throw new SAXException( "Target in Sync has no content" );
	  		  bodystate = BODY_SYNC;
	  		}
	  		break;
	  
	  case CHRS:
	  		if( bodystate == BODY_DELETE )
	  			processDelete( context,namespace,parm );
	  		else
	  		if( bodystate == BODY_RESULTS )
	  			processResults( context,namespace,parm );
	  		else
	  		if( ( bodystate == BODY_ADD ) || ( bodystate == BODY_REPLACE ) )
	  			processAdd( context,namespace,parm );
	  		else
	  		if( bodystate == BODY_GET )
	  			processGet( context,namespace,parm );
	  		else
	  		if( bodystate == BODY_STATUS )
	  			processStatus( context,namespace,parm );
	  		else
	  		if( bodystate == BODY_SYNC_CMDID ) {
			  if( syncCmdIDString == null )
	  			syncCmdIDString = parm;
			  else
				syncCmdIDString += parm;
      		} else
	  		if( bodystate == BODY_SYNC_SOURCE ) {
			  if( syncSource == null )
	  			syncSource = parm;
			  else
				syncSource += parm;
	  		} else
	  		if( bodystate == BODY_SYNC_TARGET ) {
			  if( syncTarget == null )
	  			syncTarget = parm;
			  else
				syncTarget += parm;
			}
	  		break;
	}
  }


// ------------------------- Commands
// ------------------------- Alert command
  final static int ALERT_BODY = 103;
  final static int ALERT_CMDID = 100;
  final static int ALERT_DATA = 101;
  final static int ALERT_ITEM = 102;
  
  private int alertState;
  
  private int cmdID;
  private String cmdIDString;
  private String alert_Data;
  private String alert_Source;
  private String alert_Target;
  private long alert_Last;
  private long alert_Next;
  
  private void initAlertVariables() {
    alertState = ALERT_BODY;
    cmdID = -1;
	cmdIDString = null;
    alert_Data = null;
    alert_Last = -1;
    alert_Next = -1;
    alert_Source = null;
    alert_Target = null;
  }

  private void processAlert( int context,String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
      		if( alertState == ALERT_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "CmdID" ) ) {
      		  if( alertState != ALERT_BODY )
      		  	throw new SAXException( "<CmdID> must be child of <Alert>" );
      		  alertState = ALERT_CMDID;
      		} else
      		if( parm.equals( "Data" ) ) {
      		  if( alertState != ALERT_BODY )
      		  	throw new SAXException( "<Data> must be child of <Alert>" );
      		  alertState = ALERT_DATA;
      		} else
      		if( parm.equals( "Item" ) ) {
      		  if( alertState != ALERT_BODY )
      		  	throw new SAXException( "<Item> must be child of <Alert>" );
      		  alertState = ALERT_ITEM;
      		  initItemVariables();
      		} else
      			throw new SAXException( "Illegal tag in <Alert>: "+namespace+"#"+parm );
      		break;
      		
      case END:
            if( parm.equals( "Item" ) ) {
              alertState = ALERT_BODY;
              alert_Last = item_Last;
              alert_Next = item_Next;
              alert_Source = item_Source;
              alert_Target = item_Target;
            } else
            if( alertState == ALERT_ITEM )
            	processItem( context,namespace,parm );
            else
      		if( parm.equals( "CmdID" ) ) {
      		  try {
      		    cmdID = Integer.parseInt( cmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+cmdIDString+" is not a number" );
      		  }
      		  alertState = ALERT_BODY;
			} else
			if( parm.equals( "Data" ) ) {
      		  alertState = ALERT_BODY;
      		}
      		break;
      		
      case CHRS:
      		if( alertState == ALERT_ITEM )
				processItem( context,namespace,parm );      		  
      	    else
      		if( alertState == ALERT_CMDID ) {
			  if( cmdIDString == null )
					cmdIDString = parm;
			  else
					cmdIDString += parm;
      		} else
      		if( alertState == ALERT_DATA ) {
			  if( alert_Data == null )
					alert_Data = parm;
			  else
					alert_Data += parm;
      		}
      		break;
    }
  }
  
// ------------------------------- Put command

  final static int PUT_BODY = 103;
  final static int PUT_CMDID = 100;
  final static int PUT_META = 101;
  final static int PUT_ITEM = 102;
  
  private int putState;
  
  private String put_Data;
  private String put_Source;
  private String put_Target;
  private String put_Type;

  private void initPutVariables() {
    putState = PUT_BODY;
    cmdID = -1;
	cmdIDString = null;
    put_Data = null;
    put_Source = null;
    put_Target = null;
    put_Type = null;
  }
  
  private void processPut( int context,String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
      		if( putState == PUT_META )
      			processMeta( context,namespace,parm );
      		else
      		if( putState == PUT_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "Item" ) ) {
      		  if( putState != PUT_BODY )
      		  	throw new SAXException( "<Item> must be child of <Put>" );
      		  putState = PUT_ITEM;
      		  initItemVariables();
      		} else
      		if( parm.equals( "Meta" ) ) {
      		  if( putState != PUT_BODY )
      		  	throw new SAXException( "<Meta> must be child of <Put>" );
      		  putState = PUT_META;
      		  initMetaVariables();
      		} else
      		if( parm.equals( "CmdID" ) ) {
      		  if( putState != PUT_BODY )
      		  	throw new SAXException( "<CmdID> must be child of <Put>" );
      		  putState = PUT_CMDID;
      		} else
      			throw new SAXException( "Illegal tag in <Put>: "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "Meta" ) ) {
      		  putState = PUT_BODY;
      		  if( meta_Type == null )
      		  	throw new SAXException( "<Type> not present in <Meta> inside <Put> tag" ); 
      		  put_Type = meta_Type;
      		} else
      		if( putState == PUT_META )
      			processMeta( context,namespace,parm );
      		else
            if( parm.equals( "Item" ) ) {
              putState = PUT_BODY;
              put_Data = item_Data;
              put_Source = item_Source;
            } else
            if( putState == PUT_ITEM )
            	processItem( context,namespace,parm );
            else
      		if( parm.equals( "CmdID" ) ) {
      		  try {
      		    cmdID = Integer.parseInt( cmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+cmdIDString+" is not a number" );
      		  }
      		  putState = PUT_BODY;
      		}
      		break;
      		
      case CHRS:
      		if( putState == PUT_META )
      			processMeta( context,namespace,parm );
      		else
      		if( putState == PUT_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( putState == PUT_CMDID ) {
			  if( cmdIDString == null )
				cmdIDString = parm;
			  else
				cmdIDString += parm;
      		}
      		break;
    }
  }
  
// ----------------------- Get command
  final static int GET_BODY = 103;
  final static int GET_CMDID = 100;
  final static int GET_META = 101;
  final static int GET_ITEM = 102;
  
  private int getState;
  
  private String get_Data;
  private String get_Source;
  private String get_Target;
  private String get_Type;

  private void initGetVariables() {
    getState = GET_BODY;
    cmdID = -1;
	cmdIDString = null;
    get_Data = null;
    get_Source = null;
    get_Target = null;
    get_Type = null;
  }
  
  private void processGet( int context,String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
      		if( getState == GET_META )
      			processMeta( context,namespace,parm );
      		else
      		if( getState == GET_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "Item" ) ) {
      		  if( getState != GET_BODY )
      		  	throw new SAXException( "<Item> must be child of <Get>" );
      		  getState = GET_ITEM;
      		  initItemVariables();
      		} else
      		if( parm.equals( "Meta" ) ) {
      		  if( getState != GET_BODY )
      		  	throw new SAXException( "<Meta> must be child of <Get>" );
      		  getState = GET_META;
      		  initMetaVariables();
      		} else
      		if( parm.equals( "CmdID" ) ) {
      		  if( getState != GET_BODY )
      		  	throw new SAXException( "<CmdID> must be child of <Get>" );
      		  getState = GET_CMDID;
      		} else
      			throw new SAXException( "Illegal tag in <Get>: "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "Meta" ) ) {
      		  getState = GET_BODY;
      		  if( meta_Type == null )
      		  	throw new SAXException( "<Type> not present in <Meta> inside <Put> tag" ); 
      		  get_Type = meta_Type;
      		} else
      		if( getState == GET_META )
      			processMeta( context,namespace,parm );
      		else
            if( parm.equals( "Item" ) ) {
              getState = GET_BODY;
              get_Data = item_Data;
              get_Target = item_Target;
            } else
            if( getState == GET_ITEM )
            	processItem( context,namespace,parm );
            else
      		if( parm.equals( "CmdID" ) ) {
      		  try {
      		    cmdID = Integer.parseInt( cmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+cmdIDString+" is not a number" );
      		  }
      		  getState = GET_BODY;
      		}
      		break;
      		
      case CHRS:
      		if( getState == GET_META )
      			processMeta( context,namespace,parm );
      		else
      		if( getState == GET_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( getState == GET_CMDID ) {
			  if( cmdIDString == null )
				cmdIDString = parm;
			  else
				cmdIDString += parm;
      		}
      		break;
    }
  }
  
// ----------------------- Status command
  final static int STATUS_CMDID = 100;
  final static int STATUS_ITEM = 102;
  final static int STATUS_BODY = 103;
  final static int STATUS_MSGREF = 104;
  final static int STATUS_SOURCEREF = 105;
  final static int STATUS_TARGETREF = 106;
  final static int STATUS_DATA = 107;
  final static int STATUS_CMDREF = 108;
  final static int STATUS_CMD = 109;
  
  private int statusState;
  
  private int status_Data;
  private String status_DataString;
  private int status_MsgRef;
  private String status_MsgRefString = null;
  private String status_SourceRef;
  private String status_TargetRef;
  private String status_Cmd;
  private int status_cmdRef;
  private String status_cmdRefString;

  private void initStatusVariables() {
    statusState = STATUS_BODY;
    cmdID = -1;
	cmdIDString = null;
    status_Data = -1;
	status_DataString = null;
    status_MsgRef = -1;
	status_MsgRefString = null;
    status_cmdRef = -1;
	status_cmdRefString = null;
    status_SourceRef = null;
    status_TargetRef = null;
    status_Cmd = null;
  }
  
  private void processStatus( int context,String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
      		if( statusState == STATUS_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "Item" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<Item> must be child of <Status>" );
      		  statusState = STATUS_ITEM;
      		  initItemVariables();
      		} else
      		if( parm.equals( "MsgRef" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<MsgRef> must be child of <Status>" );
      		  statusState = STATUS_MSGREF;
      		} else
      		if( parm.equals( "CmdID" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<CmdID> must be child of <Status>" );
      		  statusState = STATUS_CMDID;
      		} else
      		if( parm.equals( "Cmd" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<Cmd> must be child of <Status>" );
      		  statusState = STATUS_CMD;
      		} else      		
      		if( parm.equals( "CmdRef" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<CmdRef> must be child of <Status>" );
      		  statusState = STATUS_CMDREF;
      		} else
      		if( parm.equals( "TargetRef" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<TargetRef> must be child of <Status>" );
      		  statusState = STATUS_TARGETREF;
      		} else
      		if( parm.equals( "SourceRef" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<SourceRef> must be child of <Status>" );
      		  statusState = STATUS_SOURCEREF;
      		} else
      		if( parm.equals( "Data" ) ) {
      		  if( statusState != STATUS_BODY )
      		  	throw new SAXException( "<Data> must be child of <Status>" );
      		  statusState = STATUS_DATA;
      		} else
      			throw new SAXException( "Illegal tag in <Status>: "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "Item" ) ) {
      		  statusState = STATUS_BODY;
      		} else
      		if( statusState == STATUS_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "TargetRef" ) ||
      			parm.equals( "SourceRef" ) ||
      			parm.equals( "Cmd" ) ) {
      		  statusState = STATUS_BODY;
      		} else
			if( parm.equals( "CmdRef" ) ) {
      		  try {
      		    status_cmdRef = Integer.parseInt( status_cmdRefString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdRef "+status_cmdRefString+" is not a number" );
      		  }
      		  statusState = STATUS_BODY;
			} else
			if( parm.equals( "CmdID" ) ) {
      		  try {
      		    cmdID = Integer.parseInt( cmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+cmdIDString+" is not a number" );
      		  }
      		  statusState = STATUS_BODY;
			} else
			if( parm.equals( "Data" ) ) {
      		  try {
      		    status_Data = Integer.parseInt( status_DataString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "<Data> content "+parm+" is not a number" );
      		  }
      		  statusState = STATUS_BODY;
			} else
			if( parm.equals( "MsgRef" ) ) {
      		  try {
      		    status_MsgRef = Integer.parseInt( status_MsgRefString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "MsgRef "+status_MsgRefString+" is not a number" );
      		  }
      		  statusState = STATUS_BODY;
			}
      		break;
      		
      case CHRS:
      		if( statusState == STATUS_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( statusState == STATUS_TARGETREF ) {
			  if( status_TargetRef == null )
      		  	status_TargetRef = parm;
			  else
				status_TargetRef += parm;
      		} else
      		if( statusState == STATUS_SOURCEREF ) {
			  if( status_SourceRef == null )
      		  	status_SourceRef = parm;
			  else
				status_SourceRef += parm;
      		} else
      		if( statusState == STATUS_CMD ) {
			  if( status_Cmd == null )
      		  	status_Cmd = parm;
			  else
				status_Cmd += parm;
      		} else
      		if( statusState == STATUS_CMDREF ) {
			  if( status_cmdRefString == null )
      		  	status_cmdRefString = parm;
			  else
				status_cmdRefString += parm;
      		} else
      		if( statusState == STATUS_CMDID ) {
			  if( cmdIDString == null )
				cmdIDString = parm;
			  else
				cmdIDString += parm;
      		} else
       		if( statusState == STATUS_DATA ) {
			  if( status_DataString == null )
					status_DataString = parm;
			  else
					status_DataString += parm;
      		} else
       		if( statusState == STATUS_MSGREF ) {
			  if( status_MsgRefString == null )
					status_MsgRefString = parm;
			  else
					status_MsgRefString += parm;
      		}
      		break;
    }
  }

// ----------------------- Results command
  final static int RESULTS_CMDID = 100;
  final static int RESULTS_ITEM = 102;
  final static int RESULTS_BODY = 103;
  final static int RESULTS_MSGREF = 104;
  final static int RESULTS_CMDREF = 105;
  final static int RESULTS_META = 106;
  
  private int resultsState;
  private int results_MsgRef;
  private String results_MsgRefString;
  private int results_CmdRef;
  private String results_CmdRefString;
  private String results_Type;
  private String results_Data;
  private String results_Source;
  
  private void initResultsVariables() {
    resultsState = RESULTS_BODY;
    cmdID = -1;
	cmdIDString = null;
    results_MsgRef = -1;
	results_MsgRefString = null;
    results_CmdRef = -1;
	results_CmdRefString = null;
    results_Type = null;
    results_Data = null;
    results_Source = null;
  }
  
  private void processResults( int context,String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
            if( resultsState == RESULTS_META )
            	processMeta( context,namespace,parm );
            else
            if( resultsState == RESULTS_ITEM )
            	processItem( context,namespace,parm );
            else
      		if( parm.equals( "Item" ) ) {
      		  initItemVariables();
      		  resultsState = RESULTS_ITEM;
      		} else            
      		if( parm.equals( "Meta" ) ) {
      		  initMetaVariables();
      		  resultsState = RESULTS_META;
      		} else
      		if( parm.equals( "CmdRef" ) ) {
      		  if( resultsState != RESULTS_BODY )
      		  	throw new SAXException( "<CmdRef> must be child of <Results>" );
      		  resultsState = RESULTS_CMDREF;
      		} else
      		if( parm.equals( "MsgRef" ) ) {
      		  if( resultsState != RESULTS_BODY )
      		  	throw new SAXException( "<MsgRef> must be child of <Results>" );
      		  resultsState = RESULTS_MSGREF;
      		} else
      		if( parm.equals( "CmdID" ) ) {
      		  if( resultsState != STATUS_BODY )
      		  	throw new SAXException( "<CmdID> must be child of <Results>" );
      		  resultsState = RESULTS_CMDID;
      		} else
      			throw new SAXException( "Illegal tag in <Results>: "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "Item" ) ) {
      		  resultsState = RESULTS_BODY;
      		  if( item_Data != null )
      		  		results_Data = item_Data;
      		  if( item_Source != null )
      		  		results_Source = item_Source;
      		} else
      		if( parm.equals( "Meta" ) ) {
              resultsState = RESULTS_BODY;
              if( meta_Type != null )
              	results_Type = meta_Type;
      		} else
      		if( resultsState == RESULTS_META )
      			processMeta( context,namespace,parm );
      		else
      		if( resultsState == RESULTS_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "CmdRef" ) ) {
      		  try {
      		    results_CmdRef = Integer.parseInt( results_CmdRefString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdRef "+results_CmdRefString+" is not a number" );
      		  }
      		  resultsState = STATUS_BODY;
			}
			else
			if( parm.equals( "CmdID" ) ) {
      		  try {
      		    cmdID = Integer.parseInt( cmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+cmdIDString+" is not a number" );
      		  }
      		  resultsState = STATUS_BODY;
			} else
			if( parm.equals( "MsgRef" ) ) {
      		  try {
      		    results_MsgRef = Integer.parseInt( results_MsgRefString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "MsgRef "+results_MsgRefString+" is not a number" );
      		  }
      		  resultsState = STATUS_BODY;
			}
      		break;
      		
      case CHRS:
      		if( resultsState == RESULTS_META )
      				processMeta( context,namespace,parm );
      		else
      		if( resultsState == RESULTS_ITEM )
      				processItem( context,namespace,parm );
      		else
      		if( resultsState == RESULTS_CMDID ) {
			  if( cmdIDString == null )
				cmdIDString = parm;
			  else
				cmdIDString += parm;
      		} else
       		if( resultsState == RESULTS_MSGREF ) {
			  if( results_MsgRefString == null )
					results_MsgRefString = parm;
			  else
					results_MsgRefString += parm;
      		} else
       		if( resultsState == RESULTS_CMDREF ) {
			  if( results_CmdRefString == null )
				results_CmdRefString = parm;
			  else
				results_CmdRefString += parm;
      		}
      		break;
    }
  }
  
// ----------------------- Add/Replace command
  final static int ADD_CMDID = 100;
  final static int ADD_ITEM = 102;
  final static int ADD_BODY = 103;
  final static int ADD_META = 104;
  
  private int addState;
  private String add_Type;
  private String add_Data;
  private String add_Source;
  private String add_tagName;
  
  private void initAddVariables( String tagName ) {
    addState = ADD_BODY;
    cmdID = -1;
	cmdIDString = null;
    add_Type = null;
    add_Data = null;
    add_Source = null;
    add_tagName = tagName;
  }
  
  private void processAdd( int context,String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
      		if( addState == ADD_META )
      			processMeta( context,namespace,parm );
      		else
      		if( addState == ADD_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "Item" ) ) {
      		  if( addState != ADD_BODY )
      		  	throw new SAXException( "<Item> must be child of <"+add_tagName+">" );
      		  addState = ADD_ITEM;
      		  initItemVariables();
      		} else
      		if( parm.equals( "Meta" ) ) {
      		  if( addState != ADD_BODY )
      		  	throw new SAXException( "<Meta> must be child of <"+add_tagName+">" );
      		  addState = ADD_META;
      		  initMetaVariables();
      		} else
      		if( parm.equals( "CmdID" ) ) {
      		  if( addState != ADD_BODY )
      		  	throw new SAXException( "<CmdID> must be child of <"+add_tagName+">" );
      		  addState = ADD_CMDID;
      		} else
      			throw new SAXException( "Illegal tag in <"+add_tagName+">: "+namespace+"#"+parm );
      		break;
      	
      case END:
      		if( parm.equals( "Item" ) ) {
      		  if( item_Source != null )
      		  	add_Source = item_Source;
      		  if( item_Data != null )
      		  	add_Data = item_Data;
      		  addState = ADD_BODY;
      		} else
      		if( addState == ADD_ITEM )
      			processItem( context,namespace,parm );
            else
      		if( parm.equals( "Meta" ) ) {
      		  if( meta_Type != null )
      		  	add_Type = meta_Type;
      		  addState = ADD_BODY;
      		} else
      		if( addState == ADD_META )
      			processMeta( context,namespace,parm );
      		else
      		if( parm.equals( "CmdID" ) ) {
      		  try {
      		    cmdID = Integer.parseInt( cmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+cmdIDString+" is not a number" );
      		  }
      		  addState = ADD_BODY;
			}
      		break;
      	
      case CHRS:
      		if( addState == ADD_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( addState == ADD_META )
      			processMeta( context,namespace,parm );
      		else
      		if( addState == ADD_CMDID ) {
			  if( cmdIDString == null )
				cmdIDString = parm;
			  else
				cmdIDString += parm;
      		}
      		break;
    }
  }

// ----------------------- Delete command
  final static int DELETE_CMDID = 100;
  final static int DELETE_ITEM = 102;
  final static int DELETE_BODY = 103;
  
  private int deleteState;
  private String delete_Target;
  
  private void initDeleteVariables() {
    deleteState = DELETE_BODY;
    cmdID = -1;
	cmdIDString = null;
    delete_Target = null;
  }
  
  private void processDelete( int context,String namespace, String parm ) throws SAXException {
    switch( context ) {
      case START:
      		if( deleteState == DELETE_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( parm.equals( "Item" ) ) {
      		  if( deleteState != DELETE_BODY )
      		  	throw new SAXException( "<Item> must be child of <Delete>" );
      		  deleteState = DELETE_ITEM;
      		  initItemVariables();
      		} else
      		if( parm.equals( "CmdID" ) ) {
      		  if( deleteState != DELETE_BODY )
      		  	throw new SAXException( "<CmdID> must be child of <Delete>" );
      		  deleteState = DELETE_CMDID;
      		} else
      			throw new SAXException( "Illegal tag in <Delete>: "+namespace+"#"+parm );
      		break;
      		
      case END:
      		if( parm.equals( "Item" ) ) {
      		  if( item_Target != null )
      		  	delete_Target = item_Target;
      		  deleteState = DELETE_BODY;
      		} else
      		if( deleteState == DELETE_ITEM )
      			processItem( context,namespace,parm );
            else
      		if( parm.equals( "CmdID" ) ) {
      		  try {
      		    cmdID = Integer.parseInt( cmdIDString );
      		  } catch( NumberFormatException ex ) {
      		    throw new SAXException( "CmdID "+cmdIDString+" is not a number" );
      		  }
      		  deleteState = DELETE_BODY;
			}
      		break;
      		
      case CHRS:
      		if( deleteState == DELETE_ITEM )
      			processItem( context,namespace,parm );
      		else
      		if( deleteState == DELETE_CMDID ) {
			  if( cmdIDString == null )
				cmdIDString = parm;
			  else
				cmdIDString += parm;
      		}
      		break;
    }
  }
}
