package mwfw.sync;

import java.util.Map;
import java.util.List;
import java.io.InputStream;

/**
  * This interface is implemented by data synchronization protocol engines. 
  */
public interface ISyncProtocolAdapter {

/**
  * Opens synchronization session to a given peer for synchronizing the given
  * database.
  * @param peer URI of the peer to synchronize with. The format of this string depends
  *        of the installed protocol adapter.
  * @param syncparams Map of synchronization parameters. The content of the map
  *			depends on the sync engine, normally it includes the name of the database
  *			to synchronize and synchronization anchors (e.g. timestamp) 
  * @param cb Callback so that this component can notify the sync engine. Coding it with
  *        whiteboard pattern would create circles in the service dependency graph. 
  * @return integer handle identifying the session.
  */
  public int openSession( String peer, Map syncparams, ISyncProtocolAdapterCallBack cb ) throws Exception;

/**
  * Closes the synchronization session.
  * @param handle Handle identifying the session
  */
  public void abortSession( int handle ) throws Exception;

/**
  * Suspends the synchronization on this component. The method call will not return
  * until the component is in stable state ready for reconfiguration (e.g. to be connected with
  * another protocol adapter). The state of the ongoing sessions will remain
  * stored in the component and the sessions will be restored 
  */
  public void suspend() throws Exception;
  
/**
  * Resumes the synchronization on this component. 
  */
  public void resume() throws Exception;
  
  
/**
  * Sends a query
  * @param sessionKey session key
  * @param metainf Meta info associated with the query
  * <ul>
  * <li>type: MIME type of the data item
  * </ul>
  * @param key Absolute URI (database URI+key within the database) of the data
  * @return query handle that may be used to identify query results (returned by sendResults)
  */
  public int sendQuery( String sessionKey, Map metainf, String key ) throws Exception;
      
/**
  * Sends data addition operation
  * @param sessionKey Session key
  * @param metinf Meta information
  * <ul>
  * <li>type: MIME type of the data item
  * </ul>
  * @param key Key to write data 
  * @param is InputStream that can be used to read the data
  */
  public void sendAddition( String sessionKey, Map metinf, String key, InputStream is ) throws Exception;

/**
  * Sends data replacement operation
  * @param sessionKey Session key
  * @param metinf Meta information
  * <ul>
  * <li>type: MIME type of the data item
  * </ul>
  * @param key Key to write data 
  * @param is InputStream that can be used to read the data
  */
  public void sendReplacement( String sessionKey, Map metinf, String key, InputStream is ) throws Exception;

/**
  * Sends data deletion operation
  * @param sessionKey Session key
  * @param key Key to deleye
  */
  public void sendDeletion( String sessionKey, String key ) throws Exception;
}
