package mwfw.syncmlds.impl;

import org.xml.sax.SAXException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.InputStream;
import java.util.Map;
import java.util.Iterator;

/**
  * Utility methods used commonly in SyncML DS implementations
  */

public class SyncMLUtil {
  public static final int SYNCTYPE_TWO_WAY = 200;
  public static final int SYNCTYPE_SLOW_SYNC = 201;
  public static final int SYNCTYPE_ONE_WAY_FROM_CLIENT = 202;
  public static final int SYNCTYPE_REFRESH_FROM_CLIENT = 203;
  public static final int SYNCTYPE_ONE_WAY_FROM_SERVER = 204;
  public static final int SYNCTYPE_REFRESH_FROM_SERVER = 205;

/**
  * Sends the SyncML DS header.
  * @param sessionID ID of the SyncML session
  * @param msgID Message ID
  * @param source Sync source
  * @param target Sync target
  * @param syncparams Synchronization parameters, e.g. authentication
  *        Interpreted by this method: userid,password,maxmsgsize.
  * @param ps PrintStream to write the header into.
  */
  public static void sendSyncMLHeader( int sessionID, int msgID, 
  											String source, String target,
  											Map syncparams, PrintStream ps ) throws Exception {
    ps.println( "<SyncHdr><VerDTD>1.1</VerDTD><VerProto>SyncML/1.1</VerProto>" );
    ps.println( "<SessionID>"+sessionID+"</SessionID><MsgID>"+msgID+"</MsgID>" );
    ps.println( "<Target>"+target+"</Target><Source>"+source+"</Source>" );
    String userid = null;
    String password = null;
    String maxmsgsize = null;
    if( syncparams != null ) {
      userid = (String)syncparams.get( "userid" );
      password = (String)syncparams.get( "password" );
      maxmsgsize = (String)syncparams.get( "maxmsgsize" );
    }
    if( ( userid != null ) && ( password != null ) ) {
      String bauth = userid + ":" + password;
      String b64bauth = Base64.encodeBytes( bauth.getBytes( "UTF-8" ) );
      ps.println( "<Cred><Meta><Type xmlns='syncml:metinf'>syncml:auth-basic</Type></Meta>" );
      ps.println( "<Data>"+b64bauth+"</Data></Cred>" );
    }
    if( maxmsgsize != null ) {
      ps.println( "<Meta><MaxMsgSize xmlns='syncml:metinf'>"+maxmsgsize+"</MaxMsgSize></Meta>" );
    }
    ps.println( "</SyncHdr>" ); 
  }
  
/**
  * Sends characteristics for one database for the Alert in Msg1
  * @param cmdID CmdID of the Alert commanc
  * @param syncType Synchronization type (see SYNCTYPE* constants)
  * @param last Last synchronization's counter
  * @param next This synchronization's counter
  * @param source Source database URI
  * @param target Target database URI
  * @param s PrintStream to write the SyncML fragment to.
  */
  public static void sendDbAlert( int cmdID, int syncType, 
  									long last,long next, 
  									String source, String target,
  									PrintStream ps ) throws Exception {
    ps.println( "<Alert><CmdID>"+cmdID+"</CmdID><Data>"+syncType+"</Data>" );
    ps.println( "<Item><Target><LocURI>"+target+"</LocURI></Target>" );
    ps.println( "<Source><LocURI>"+source+"</LocURI></Source>" );
    ps.println( "<Meta><Anchor xmlns='syncml:metinf'><Last>"+last+"</Last>" );
    ps.println( "<Next>"+next+"</Next></Anchor></Meta></Item></Alert>" );
  }
  
/**
  * Sends Devinfo in Pkg#1
  * @param cmdID CmdID of the Put command. The Get command will be issued with cmdID+1
  * @param devinfo key-value pairs of the devinfo values
  * @param s PrintStream to write the result into.
  */
  public static void sendDevInfo( int cmdId, Map devinfo, PrintStream ps ) throws Exception {
    if( devinfo != null ) {
      ps.println( "<Put><CmdID>"+cmdId+"</CmdID>" );
      ps.println( "<Meta><Type xmlns='syncml:metinf'>application/vnd.syncml-devinf+xml</Type></Meta>" );
      ps.println( "<Item><Source><LocURI>./devinf11</LocURI></Source>" );
      ps.println( "<Data><DevInf xmlns='syncml:devinf'>" );
      Iterator i = devinfo.keySet().iterator();
      for( ; i.hasNext() ; ) {
        String devinfokey = (String)i.next();
        String devinfoval = (String)devinfo.get( (Object)devinfokey );
        ps.println( "<"+devinfokey+">"+devinfoval+"</"+devinfokey+">" );
      }
      ps.println( "</DevInf></Data></Item></Put>" );
    }
    cmdId++;
    ps.println( "<Get><CmdID>"+cmdId+"</CmdID>" );
    ps.println( "<Meta><Type xmlns='syncml:metinf'>application/vnd.syncml-devinf+xml</Type></Meta>" );
    ps.println( "<Item><Target><LocURI>./devinf11</LocURI></Target></Item></Get>" );
  }
  

  public static void sendGetCommand( int cmdId, String type, String target, PrintStream ps ) throws SAXException {
    try {
      ps.println( "<Get><CmdID>"+cmdId+"</CmdID>" );
      ps.println( "<Meta><Type xmlns='syncml:metinf'>"+type+"</Type></Meta>" );
      ps.println( "<Item><Target><LocURI>"+target+"</LocURI></Target></Item></Get>" );
    } catch( Exception ex ) {
      throw new SAXException( ex );
    }
  }

/**
  * Checks whether message header needs to be sent for the session response message and
  * sends it if necessary
  */
  public static void sendResponseHeader( SyncSession session, PrintStream ps ) throws SAXException {
    if( session.responseHeaderForMsg < session.lastMsgID ) {
      session.responseHeaderForMsg = session.lastMsgID;
      try {
// As this is a response, target and source are swapped compared to the request
        sendSyncMLHeader( session.sessionID, session.lastMsgID, 
  											session.target, session.source,
  											null, ps );
  	  } catch( Exception ex ) {
  	    ex.printStackTrace();
  	    throw new SAXException( ex );
  	  }
    }
  }
  
/**
  * Sends a status 
  * @param cmdID cmdID of the status command.
  * @param msgRef Reference of the message this Status refers to
  * @param cmdIDRef cmdID of the command this Status refers to
  * @param cmdRef Name of the command this Status refers to
  * @param srcRef SourceRef tag value
  * @param targetRef TargetRef tag value
  * @param statusCode status code in Data tag
  * @param misc Additional tags in Status tag (may be null)
  * @param ps PrintStream to write the response into
  */
  public static void sendStatus( int cmdID, int msgRef, int cmdIDRef, String cmdRef,
  								String srcRef, String targetRef,
  								int statusCode, String misc, PrintStream ps ) throws SAXException {
  	try {
      ps.println( "<Status><CmdID>"+cmdID+"</CmdID>" );
      ps.println( "<MsgRef>"+msgRef+"</MsgRef><CmdRef>"+cmdIDRef+"</CmdRef><Cmd>"+cmdRef+"</Cmd>" );
      if( targetRef != null )
      	ps.println( "<TargetRef>"+targetRef+"</TargetRef>" );
      if( srcRef != null )
      	ps.println( "<SourceRef>"+srcRef+"</SourceRef>" );
      ps.println( "<Data>"+statusCode+"</Data>" );
      if( misc != null )
    	ps.println( misc ); 
      ps.println( "</Status>" );
    } catch( Exception ex ) {
      throw new SAXException( ex );
    }
  }
  
/**
  * Sends a result
  * @param cmdID CmdID of the Results tag
  * @param msgRef The message this Results refers to
  * @param cmdRef The command this Results refers to
  * @param type MIME type of the data item this Results carries
  * @parma source Source of this data item
  * @param is InputStream to read the data item
  * @param item Alternatively, the data item can be specified as a string. Either is or item
  *             may be null.
  * @param ps PrintStream to write the data to
  */
  public static void sendResults( int cmdID, int msgRef, int cmdRef, String type, String source, 
  									InputStream is,String item, PrintStream ps ) throws SAXException {
    try {
      ps.println( "<Results><CmdID>"+cmdID+"</CmdID><MsgRef>"+msgRef+"</MsgRef><CmdRef>"+cmdRef+"</CmdRef>" );
      ps.println( "<Meta><Type xmlns='syncml:metinf'>"+type+"</Type></Meta>" );
      ps.println( "<Item><Source><LocURI>"+source+"</LocURI></Source><Data>" );
      if( item != null )
      	ps.print( item );
      else
      if( is != null ) {
        int c = 0;
        do {
          c = is.read();
          if( c >= 0 )
          	ps.print( (char)c );
        } while( c >= 0 );
      }
      ps.println( "</Data></Item></Results>" );
    } catch( Exception ex ) {
      throw new SAXException( ex );
    }
  }
  
/**
  * Sends a general Alert.
  * @param cmdID cmdID of the Alert command
  * @param data Alert type
  * @param source Alert source
  * @param target Alert target
  * @param tags Additional elements in Alert body (may be null)
  * @param ps PrintStream to write the result to
  */
  public static void sendAlert( int cmdID, int data, String source, String target, 
  							String tags, PrintStream ps ) throws SAXException {
    try {
      ps.println( "<Alert><CmdID>"+cmdID+"</CmdID><Data>"+data+"</Data>" );
      ps.println( "<Item><Target><LocURI>"+target+"</LocURI></Target>" );
      ps.println( "<Source><LocURI>"+source+"</LocURI></Source>" );
      if( tags != null )
      	ps.println( tags );
      ps.println( "</Item></Alert>" );
    } catch( Exception ex ) {
      throw new SAXException( ex );
    }
  }

/**
  * Sends an Add command
  * @param cmdID cmdID of the Add command
  * @param source Source of the Add tag (source key of the data)
  * @param type MIME type of the 
  * @param ps PrintStream to write the result to
  */
  public static void sendAddCommand( int cmdID, String source,  
  							String type, InputStream data, PrintStream ps ) throws SAXException {
    try {
      ps.println( "<Add><CmdID>"+cmdID+"</CmdID>" );
      ps.println( "<Meta><Type xmlns='syncml:metinf'>"+type+"</Type></Meta>" );
      ps.println( "<Item><Source><LocURI>"+source+"</LocURI></Source><Data>" );
      int c = 0;
      do {
        c = data.read();
        if( c >= 0 )
        	ps.print( (char)c );
      } while( c >= 0 );
      ps.println( "</Data></Item></Add>" );
    } catch( Exception ex ) {
      ex.printStackTrace();
      throw new SAXException( ex );
    }
  }

/**
  * Sends an Replace command
  * @param cmdID cmdID of the Add command
  * @param source Source of the Add tag (source key of the data)
  * @param type MIME type of the item
  * @param data Stream to read the item from
  * @param ps PrintStream to write the result to
  */
  public static void sendReplaceCommand( int cmdID, String source,  
  							String type, InputStream data, PrintStream ps ) throws SAXException {
    try {
      ps.println( "<Replace><CmdID>"+cmdID+"</CmdID>" );
      ps.println( "<Meta><Type xmlns='syncml:metinf'>"+type+"</Type></Meta>" );
      ps.println( "<Item><Source><LocURI>"+source+"</LocURI></Source><Data>" );
      int c = 0;
      do {
        c = data.read();
        if( c >= 0 )
        	ps.print( (char)c );
      } while( c >= 0 );
      ps.println( "</Data></Item></Replace>" );
    } catch( Exception ex ) {
      throw new SAXException( ex );
    }
  }

/**
  * Sends a Delete command
  * @param cmdID cmdID of the Add command
  * @param target Key to delete
  * @param ps PrintStream to write the result to
  */
  public static void sendDeleteCommand( int cmdID, String target, PrintStream ps ) throws SAXException {
    try {
      ps.println( "<Delete><CmdID>"+cmdID+"</CmdID>" );
      ps.println( "<Item><Target><LocURI>"+target+"</LocURI></Target></Item>" );
      ps.println( "</Delete>" );
    } catch( Exception ex ) {
      throw new SAXException( ex );
    }
  }
}
