package mwfw.sync;

import java.util.HashMap;
import java.util.Vector;
import java.util.Map;
import java.util.ArrayList;
import java.util.Dictionary;
import java.io.InputStream;
import java.util.Iterator;

class TimeStampSession {
  String target = null;
  String target_db = null;
  long last = -1;
  long next = -1;
  IStorage storage = null;
  boolean client = false;
}

/**
  * Time-stamp-based synchronization engine.
  * The component must have mwfw.sync.timestamp.source property set
  */

public class TimeStampSyncEngine implements ISyncControl,ISyncProtocolAdapterCallBack {
  HashMap busy = new HashMap();
  Vector prepref = new Vector();
  boolean active = false;
  ISyncProtocolAdapter adapt = null;
  String source = null;
  HashMap sessions = new HashMap();
  ArrayList listeners = new ArrayList();
  HashMap<String,IStorage> storage = new HashMap<String,IStorage>();

  public void init( HashMap configMap ) throws RuntimeException {
    active = true;
    source = (String)configMap.get( "mwfw.sync.timestamp.source" );
    if( source == null )
    	throw new RuntimeException( "mwfw.sync.timestamp.source not set for TimeStampSyncEngine" );
  }

  public void shutdown() {
    active = false;
	storage = null;
    busy.clear();
    sessions.clear();
  }

  public void setStorage( String dbName, IStorage db ) {
	storage.put( dbName, db );
  }
  
  public void unsetStorage( String dbName ) {
	storage.remove( dbName );
  }
  
  public void setSyncProtocolAdapter( ISyncProtocolAdapter adapt ) {
    this.adapt = adapt;
  }

  public void unsetSyncProtocolAdapter() {
    this.adapt = null;
  }

// ISyncControl
  public void startSync( String peer, String database ) throws Exception {
    if( adapt == null )
    	throw new RuntimeException( "Sync protocol adapter not connected" );
    IStorage is = storage.get( database );
    if( is == null )
    	throw new RuntimeException( "Database URI not registered: "+database );
    Map m = is.getDatabaseMetaInfo();
    long last = 0L;
    if( m != null ) {
      String key = peer + "_last";
      String ls = (String)m.get( key );
      try {
        last = Long.parseLong( ls );
      } catch( NumberFormatException ex ) {}
    }
    HashMap syncparams = new HashMap();
    syncparams.put( "source", source );
    syncparams.put( "dbname", database );
    syncparams.put( "last", new Long( last ) );
    syncparams.put( "next", new Long( System.currentTimeMillis() ) );
    adapt.openSession( peer, syncparams, (ISyncProtocolAdapterCallBack)this );
  }

  public void addSyncListener( ISyncListener synclistener ) {
    synchronized( listeners ) {
      if( !listeners.contains( synclistener ) )
      	listeners.add( synclistener );
    }
  }

  public void removeSyncListener( ISyncListener synclistener ) {
    synchronized( listeners ) {
      int idx = listeners.indexOf( synclistener );
      if( idx >= 0 )
      	listeners.remove( idx );
    }
  }

  private void notifySyncListeners( String sessionKey, boolean endSync, boolean client ) {
    synchronized( listeners ) {
      for( int i = 0 ; i < listeners.size() ; ++i ) {
        ISyncListener isl = (ISyncListener)listeners.get( i );
        try {
          if( endSync )
        	isl.sessionEnded( sessionKey,client );
          else
        	isl.sessionStarted( sessionKey, client );
        } catch( Exception ex ) {
          ex.printStackTrace();
        }
      }
    }
  }
// ISyncProtocolAdapterCallBack

  public void syncMetaInfo( String sessionKey, Map metainf, boolean client ) throws Exception {
    String target = (String)metainf.get( "target" );
    if( target == null )
    	throw new RuntimeException( "target not set" );
    String target_db = (String)metainf.get( "target_db" );
    if( target_db == null )
    	throw new RuntimeException( "target_db not set" );
    Long lastl = (Long)metainf.get( "last" );
    if( lastl == null )
    	throw new RuntimeException( "last not set" );
    Long nextl = (Long)metainf.get( "next" );
    if( nextl == null )
    	throw new RuntimeException( "next not set" );
    IStorage is = storage.get( target_db );
    if( is == null )
    	throw new RuntimeException( "Database URI not registered: "+target_db );
    if( !is.lock( sessionKey ) )
    	throw new RuntimeException( "Database "+target_db+" cannot be locked for session key: "+sessionKey );
    TimeStampSession session = new TimeStampSession();
    session.target = target;
    session.target_db = target_db;
    session.last = lastl.longValue();
    session.next = nextl.longValue();
    session.storage = is;
    session.client = client;
    synchronized( sessions ) {
      sessions.put( sessionKey, session );
    }
    notifySyncListeners( sessionKey, false, client );
//    System.out.println( "syncMetaInfo: app: "+appId+" engine: "+engineId+"source: "+source+" sessionKey: "+sessionKey );
  }
  
  public InputStream syncGet( String sessionKey, Map metainf, String key ) {
    return null;
  }

  public void syncQueryResult( String sessionKey, int handle, 
					Map metainf, String key, InputStream is ) {
  } 

  public void sendCommands( String sessionKey,ISyncProtocolAdapter adapt ) throws Exception {
    TimeStampSession session = (TimeStampSession)sessions.get( sessionKey );
    if( session == null )
    	throw new RuntimeException( "no sync engine session for session key: "+sessionKey );
    IStorage storage = session.storage;
	HashMap queryMap = new HashMap();
	queryMap.put( "timestampgeq",new Long( session.last+1 ) );    
    Iterator i = storage.selectKeys( queryMap );
    for( ; i.hasNext() ; ) {
      String key = (String)i.next();
      Map entrymetainf = storage.getEntryMetaInfo( key );
      long deleted = -1;
      try {
        deleted = Long.parseLong( (String)entrymetainf.get( "deleted" ) );
      } catch( Exception ex ) {}
      long created = -1;
      try {
         created = Long.parseLong( (String)entrymetainf.get( "created" ) );
      } catch( Exception ex ) {}
      long timestamp = -1;
      try {
         timestamp = Long.parseLong( (String)entrymetainf.get( "timestamp" ) );
      } catch( Exception ex ) {}
	  System.out.println( "sendCommands: key: "+key+
						",deleted: "+deleted+
						",created: "+created+
						",timestamp: "+timestamp );
      if( created > session.last ) {
        if( adapt != null ) {
          String type = (String)entrymetainf.get( "type" );
      	  if( type == null )
      		continue;
          HashMap addMetaInf = new HashMap();
          addMetaInf.put( "type",type );
          InputStream is = storage.getEntryData( key );
          if( is == null )
      			continue;
          adapt.sendAddition( sessionKey, addMetaInf, key, is );
        }
      } else
      if( deleted > session.last ) {
        if( adapt != null ) {
          adapt.sendDeletion( sessionKey, key );
        }
      } else
      if ( timestamp > session.last ) { // condition check just for paranoia ...
        if( adapt != null ) {
          String type = (String)entrymetainf.get( "type" );
      	  if( type == null )
      		throw new RuntimeException( "no type for key: "+key );
          HashMap addMetaInf = new HashMap();
          addMetaInf.put( "type",type );
          InputStream is = storage.getEntryData( key );
          if( is == null )
      			throw new RuntimeException( "no data for key: "+key );
          adapt.sendReplacement( sessionKey, addMetaInf, key, is );
        }        
      }      
    }
  }

  public void syncAddReplace( String sessionKey, boolean replace, Map metainf, 
  							String source_db, String target_db, String key, InputStream is ) throws Exception {
    TimeStampSession session = (TimeStampSession)sessions.get( sessionKey );
    if( session == null )
    	throw new RuntimeException( "no sync engine session for session key: "+sessionKey );
    IStorage storage = session.storage;
    String type = (String)metainf.get( "type" );
    if( type == null )
    	throw new RuntimeException( "no type information for incoming request for key: "+key );
    Map entrymetainf = null;    
  	if( replace ) { // Replace
      entrymetainf = storage.getEntryMetaInfo( key );
      if( entrymetainf == null )
    	throw new RuntimeException( "no metainfo for the entry of key "+key+" (session key: "+sessionKey+")" );
  	  entrymetainf.put( "type",type );
  	  entrymetainf.remove( "deleted" );
  	  entrymetainf.put( "timestamp",Long.toString( session.next ) );
  	  storage.setEntryData( key, entrymetainf, is );
  	} else { // Add
  	  entrymetainf = (Map)new HashMap();
  	  entrymetainf.put( "type",type );
  	  entrymetainf.remove( "deleted" );
  	  String ts = Long.toString( session.next );
  	  entrymetainf.put( "timestamp",ts );
  	  entrymetainf.put( "created",ts );
  	  storage.setEntryData( key, entrymetainf, is );  	  
  	}
  }

  public void syncDelete( String sessionKey, String target_db, String key ) throws Exception {
    TimeStampSession session = (TimeStampSession)sessions.get( sessionKey );
    if( session == null )
    	throw new RuntimeException( "no sync engine session for session key: "+sessionKey );
    IStorage storage = session.storage;
    HashMap metainf = new HashMap();
  	String ts = Long.toString( session.next );
    metainf.put( "deleted",ts );
    metainf.put( "timestamp",ts );
    storage.deleteEntryData( key, metainf );
  }

  public void endSession( String sessionKey ) throws Exception {
    //System.out.println( "endSession: app: "+appId+" engine: "+engineId+" source: "+
    //					source+" sessionKey: "+sessionKey );
    TimeStampSession session = (TimeStampSession)sessions.get( sessionKey );
    if( session == null )
    	throw new RuntimeException( "no sync engine session for session key: "+sessionKey );
    IStorage storage = session.storage;
    storage.commit( sessionKey );
    Map dbmetainf = storage.getDatabaseMetaInfo();
    String key = session.target + "_last";
    dbmetainf.put( key,Long.toString( session.next ) );
    storage.setDatabaseMetaInfo( dbmetainf );    
    synchronized( sessions ) {
      sessions.remove( sessionKey );
    }
    notifySyncListeners( sessionKey, true, session.client );
  }
  

}
