package mwfw.syncmlds.impl;

import org.xml.sax.SAXException;
import java.io.InputStream;

/**
  * This interface is implemented by classes that want to receive notifications about parsed
  * methods from the SyncML DS package parser.
  */
public interface IReceiveCommand {

/**
  * End of SyncHdr. This callback may be used to retrive SyncHdr information from ParserState instance
  * @param parser Parser object processing the header
  */
  public void endSyncHdr( ParserState parser ) throws SAXException;

/**
  * Database alert in Pkg#1
  * @param Unique key of this session
  * @param cmdID Value of the CmdID structure
  * @param type Sync type, check SyncMLUtil.SYNCTYPE*
  * @param source Source database to synchronize
  * @param target Destination database to synchronize
  * @param last Last timestamp (timestamp when the database was synchronized last time)
  * @param next Next timestamp (timestamp of this synchronization session)
  */
  public void dbAlertInPackage1( String sessionKey, int cmdID, int type, String source, String target, 
  							long last, long next ) throws SAXException;
  
/**
  * Put command in package 1 conveying devinfo
  * @param sessionKey Unique key of this session
  * @param cmdID Value of the CmdID structure
  * @param devinf The value of the <Item> tag as unparsed string.
  */
  public void putInPackage1( String sessionKey, int cmdID, String devinf ) throws SAXException;

/**
  * Get command
  * @param sessionKey Unique key of this session
  * @param msgID ID of the message the Get command appeared in
  * @param cmdID cmdID of the command
  * @param source Source field of the Get command
  */
  public void getCommand( String sessionKey, int msgID, int cmdID, String type, String target ) throws SAXException;

/**
  * Status command
  * @param sessionKey Unique key of this session
  * @param cmdID cmdID of this Status command
  * @param data Status code (content of the <Data> tag)
  * @param msgRef ID of the message this Status refers to
  * @param cmdRef ID of the command this status refers to
  * @param sourceRef Source value this Status refers to
  * @param targetRef Target value this Status refers to
  */
   public void statusCommand( String sessionKey, int msgID, int cmdID, int data, int msgRef, int cmdRef,
   							  String cmd, String sourceRef, String targetRef ) throws SAXException;

/**
  * Results command
  * @param sessionKey Session Key
  * @param msgID message ID
  * @param cmdID cmdID of this Status command
  * @param msgRef ID of the message this Results refers to
  * @param cmdRef ID of the command this Results refers to
  * @param results_Source Source tag of the Results command
  * @param results_Type MIME type of the result data
  * @param results_Data Data of the results command
  */
   public void resultsCommand( String sessionKey, int msgID, int cmdID, int msgRef, int cmdRef, 
      		  							String results_Source, String results_Type, 
      		  							String results_Data ) throws SAXException;

/**
  * Add/Replace command
  * @param sessionKey Session Key
  * @param replace true if this is a Replace command, false if this is an Add
  * @param msgID message ID
  * @param cmdID CmdID of this command
  * @param source_db Source database (taken from Sync command's Source element)
  * @param target_db Target database (taken from Sync command's Target element)
  * @param source Source tag of the Add/Replace command
  * @param type MIME type of the data to be added/replaced
  * @param data Data to add/replace
  */
  public void addReplaceCommand( String sessionKey, boolean replace, int msgID, int cmdID,  
  										String source_db,String target_db,
      		  							String source, String type, 
      		  							String data ) throws SAXException;

/**
  * Delete command
  * @param sessionKey session key
  * @param msgID message ID
  * @param cmdID command ID
  * @param target_db Target database in which the entry will be deleted
  * @param target ID of the entry to be deleted
  */
  public void deleteCommand( String sessionKey, int msgID, int cmdID, String target_db, String target )
  							throws SAXException;

/**
  * Sync command
  * @param sessionKey session key
  * @param msgID message ID
  * @param cmdID CmdID of this command
  * @param source_db Source database 
  * @param target_db Target database 
  */
  public void syncCommand( String sessionKey, int msgID, int cmdID, String source_db, String target_db )
  							throws SAXException;
  							
/**
  * Delivered at the end of messages
  * @param sessionKey session key
  * @param msgID Message ID
  */ 
  public void endMessage( String sessionKey, int msgID ) throws SAXException;
}
